@@ -1,57 +1,58 @@
+from __future__ import annotations
+
 import re
 from functools import partial
-from warnings import warn
 from pathlib import Path
 from types import MappingProxyType
-from typing import Callable, Type, TypeVar, Union
-from typing import Collection, Sequence, Mapping
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Literal,
+    TypeVar,
+)
+from warnings import warn
 
 import h5py
 import numpy as np
 import pandas as pd
 from scipy import sparse
 
-from .._core.sparse_dataset import SparseDataset
-from .._core.file_backing import AnnDataFileManager
+from anndata._warnings import OldFormatWarning
+
 from .._core.anndata import AnnData
+from .._core.file_backing import AnnDataFileManager, filename
+from .._core.sparse_dataset import BaseCompressedSparseDataset
 from ..compat import (
-    _from_fixed_length_strings,
-    _decode_structured_array,
     _clean_uns,
-    Literal,
+    _decode_structured_array,
+    _from_fixed_length_strings,
 )
+from ..experimental import read_dispatched
+from .specs import read_elem, write_elem
+from .specs.registry import IOSpec, write_spec
 from .utils import (
     H5PY_V3,
+    _read_legacy_raw,
+    idx_chunks_along_axis,
     report_read_key_on_error,
     report_write_key_on_error,
-    idx_chunks_along_axis,
-    _read_legacy_raw,
 )
-from .specs import read_elem, write_elem
-from anndata._warnings import OldFormatWarning
+
+if TYPE_CHECKING:
+    from collections.abc import Collection, Mapping, Sequence
 
 T = TypeVar("T")
 
 
 def write_h5ad(
-    filepath: Union[Path, str],
+    filepath: Path | str,
     adata: AnnData,
     *,
-    force_dense: bool = None,
     as_dense: Sequence[str] = (),
     dataset_kwargs: Mapping = MappingProxyType({}),
     **kwargs,
 ) -> None:
-    if force_dense is not None:
-        warn(
-            "The `force_dense` argument is deprecated. Use `as_dense` instead.",
-            FutureWarning,
-        )
-    if force_dense is True:
-        if adata.raw is not None:
-            as_dense = ("X", "raw/X")
-        else:
-            as_dense = ("X",)
     if isinstance(as_dense, str):
         as_dense = [as_dense]
     if "raw.X" in as_dense:
@@ -72,19 +73,24 @@ def write_h5ad(
     mode = "a" if adata.isbacked else "w"
     if adata.isbacked:  # close so that we can reopen below
         adata.file.close()
+
     with h5py.File(filepath, mode) as f:
         # TODO: Use spec writing system for this
+        # Currently can't use write_dispatched here because this function is also called to do an
+        # inplace update of a backed object, which would delete "/"
         f = f["/"]
         f.attrs.setdefault("encoding-type", "anndata")
         f.attrs.setdefault("encoding-version", "0.1.0")
 
-        if "X" in as_dense and isinstance(adata.X, (sparse.spmatrix, SparseDataset)):
+        if "X" in as_dense and isinstance(
+            adata.X, (sparse.spmatrix, BaseCompressedSparseDataset)
+        ):
             write_sparse_as_dense(f, "X", adata.X, dataset_kwargs=dataset_kwargs)
         elif not (adata.isbacked and Path(adata.filename) == Path(filepath)):
             # If adata.isbacked, X should already be up to date
             write_elem(f, "X", adata.X, dataset_kwargs=dataset_kwargs)
         if "raw/X" in as_dense and isinstance(
-            adata.raw.X, (sparse.spmatrix, SparseDataset)
+            adata.raw.X, (sparse.spmatrix, BaseCompressedSparseDataset)
         ):
             write_sparse_as_dense(
                 f, "raw/X", adata.raw.X, dataset_kwargs=dataset_kwargs
@@ -106,12 +112,18 @@ def write_h5ad(
 
 
 @report_write_key_on_error
-def write_sparse_as_dense(f, key, value, dataset_kwargs=MappingProxyType({})):
+@write_spec(IOSpec("array", "0.2.0"))
+def write_sparse_as_dense(
+    f: h5py.Group,
+    key: str,
+    value: sparse.spmatrix | BaseCompressedSparseDataset,
+    *,
+    dataset_kwargs: Mapping[str, Any] = MappingProxyType({}),
+):
     real_key = None  # Flag for if temporary key was used
     if key in f:
-        if (
-            isinstance(value, (h5py.Group, h5py.Dataset, SparseDataset))
-            and value.file.filename == f.file.filename
+        if isinstance(value, BaseCompressedSparseDataset) and (
+            filename(value.group) == filename(f)
         ):  # Write to temporary key before overwriting
             real_key = key
             # Transform key to temporary, e.g. raw/X -> raw/_X, or X -> _X
@@ -128,7 +140,7 @@ def write_sparse_as_dense(f, key, value, dataset_kwargs=MappingProxyType({})):
         del f[key]
 
 
-def read_h5ad_backed(filename: Union[str, Path], mode: Literal["r", "r+"]) -> AnnData:
+def read_h5ad_backed(filename: str | Path, mode: Literal["r", "r+"]) -> AnnData:
     d = dict(filename=filename, filemode=mode)
 
     f = h5py.File(filename, mode)
@@ -147,27 +159,21 @@ def read_h5ad_backed(filename: Union[str, Path], mode: Literal["r", "r+"]) -> An
 
     d["raw"] = _read_raw(f, attrs={"var", "varm"})
 
-    X_dset = f.get("X", None)
-    if X_dset is None:
-        pass
-    elif isinstance(X_dset, h5py.Group):
-        d["dtype"] = X_dset["data"].dtype
-    elif hasattr(X_dset, "dtype"):
-        d["dtype"] = f["X"].dtype
-    else:
-        raise ValueError()
+    adata = AnnData(**d)
 
-    _clean_uns(d)
+    # Backwards compat to <0.7
+    if isinstance(f["obs"], h5py.Dataset):
+        _clean_uns(adata)
 
-    return AnnData(**d)
+    return adata
 
 
 def read_h5ad(
-    filename: Union[str, Path],
-    backed: Union[Literal["r", "r+"], bool, None] = None,
+    filename: str | Path,
+    backed: Literal["r", "r+"] | bool | None = None,
     *,
     as_sparse: Sequence[str] = (),
-    as_sparse_fmt: Type[sparse.spmatrix] = sparse.csr_matrix,
+    as_sparse_fmt: type[sparse.spmatrix] = sparse.csr_matrix,
     chunk_size: int = 6000,  # TODO, probably make this 2d chunks
 ) -> AnnData:
     """\
@@ -182,6 +188,13 @@ def read_h5ad(
         instead of fully loading it into memory (`memory` mode).
         If you want to modify backed attributes of the AnnData object,
         you need to choose `'r+'`.
+
+        Currently, `backed` only support updates to `X`. That means any
+        changes to other slots like `obs` will not be written to disk in
+        `backed` mode. If you would like save changes made to these slots
+        of a `backed` :class:`~anndata.AnnData`, write them to a new file
+        (see :meth:`~anndata.AnnData.write`). For an example, see
+        [here] (https://anndata-tutorials.readthedocs.io/en/latest/getting-started.html#Partial-reading-of-large-data).
     as_sparse
         If an array was saved as dense, passing its name here will read it as
         a sparse_matrix, by chunk of size `chunk_size`.
@@ -222,45 +235,51 @@ def read_h5ad(
     )
 
     with h5py.File(filename, "r") as f:
-        d = {}
-        for k in f.keys():
-            # Backwards compat for old raw
-            if k == "raw" or k.startswith("raw."):
-                continue
-            if k == "X" and "X" in as_sparse:
-                d[k] = rdasp(f[k])
-            elif k == "raw":
-                assert False, "unexpected raw format"
-            elif k in {"obs", "var"}:
+
+        def callback(func, elem_name: str, elem, iospec):
+            if iospec.encoding_type == "anndata" or elem_name.endswith("/"):
+                return AnnData(
+                    **{
+                        # This is covering up backwards compat in the anndata initializer
+                        # In most cases we should be able to call `func(elen[k])` instead
+                        k: read_dispatched(elem[k], callback)
+                        for k in elem.keys()
+                        if not k.startswith("raw.")
+                    }
+                )
+            elif elem_name.startswith("/raw."):
+                return None
+            elif elem_name == "/X" and "X" in as_sparse:
+                return rdasp(elem)
+            elif elem_name == "/raw":
+                return _read_raw(f, as_sparse, rdasp)
+            elif elem_name in {"/obs", "/var"}:
                 # Backwards compat
-                d[k] = read_dataframe(f[k])
-            else:  # Base case
-                d[k] = read_elem(f[k])
+                return read_dataframe(elem)
+            return func(elem)
 
-        d["raw"] = _read_raw(f, as_sparse, rdasp)
+        adata = read_dispatched(f, callback=callback)
 
-        X_dset = f.get("X", None)
-        if X_dset is None:
-            pass
-        elif isinstance(X_dset, h5py.Group):
-            d["dtype"] = X_dset["data"].dtype
-        elif hasattr(X_dset, "dtype"):
-            d["dtype"] = f["X"].dtype
-        else:
-            raise ValueError()
+        # Backwards compat (should figure out which version)
+        if "raw.X" in f:
+            raw = AnnData(**_read_raw(f, as_sparse, rdasp))
+            raw.obs_names = adata.obs_names
+            adata.raw = raw
 
-    _clean_uns(d)  # backwards compat
+        # Backwards compat to <0.7
+        if isinstance(f["obs"], h5py.Dataset):
+            _clean_uns(adata)
 
-    return AnnData(**d)
+    return adata
 
 
 def _read_raw(
-    f: Union[h5py.File, AnnDataFileManager],
+    f: h5py.File | AnnDataFileManager,
     as_sparse: Collection[str] = (),
-    rdasp: Callable[[h5py.Dataset], sparse.spmatrix] = None,
+    rdasp: Callable[[h5py.Dataset], sparse.spmatrix] | None = None,
     *,
     attrs: Collection[str] = ("X", "var", "varm"),
-):
+) -> dict:
     if as_sparse:
         assert rdasp is not None, "must supply rdasp if as_sparse is supplied"
     raw = {}
@@ -274,7 +293,7 @@ def _read_raw(
 
 
 @report_read_key_on_error
-def read_dataframe_legacy(dataset) -> pd.DataFrame:
+def read_dataframe_legacy(dataset: h5py.Dataset) -> pd.DataFrame:
     """Read pre-anndata 0.7 dataframes."""
     warn(
         f"'{dataset.name}' was written with a very old version of AnnData. "
@@ -293,7 +312,7 @@ def read_dataframe_legacy(dataset) -> pd.DataFrame:
     return df
 
 
-def read_dataframe(group) -> pd.DataFrame:
+def read_dataframe(group: h5py.Group | h5py.Dataset) -> pd.DataFrame:
     """Backwards compat function"""
     if not isinstance(group, h5py.Group):
         return read_dataframe_legacy(group)
@@ -340,7 +359,7 @@ def read_dense_as_sparse(
         raise ValueError(f"Cannot read dense array as type: {sparse_format}")
 
 
-def read_dense_as_csr(dataset, axis_chunk=6000):
+def read_dense_as_csr(dataset: h5py.Dataset, axis_chunk: int = 6000):
     sub_matrices = []
     for idx in idx_chunks_along_axis(dataset.shape, 0, axis_chunk):
         dense_chunk = dataset[idx]
@@ -349,7 +368,7 @@ def read_dense_as_csr(dataset, axis_chunk=6000):
     return sparse.vstack(sub_matrices, format="csr")
 
 
-def read_dense_as_csc(dataset, axis_chunk=6000):
+def read_dense_as_csc(dataset: h5py.Dataset, axis_chunk: int = 6000):
     sub_matrices = []
     for idx in idx_chunks_along_axis(dataset.shape, 1, axis_chunk):
         sub_matrix = sparse.csc_matrix(dataset[idx])
