@@ -1,11 +1,12 @@
-from pathlib import Path
-from os import PathLike, fspath
-from types import MappingProxyType
-from typing import Union, Optional, Mapping, Tuple
-from typing import Iterable, Iterator, Generator
-from collections import OrderedDict
-import gzip
+from __future__ import annotations
+
 import bz2
+import gzip
+from collections import OrderedDict
+from os import PathLike, fspath
+from pathlib import Path
+from types import MappingProxyType
+from typing import TYPE_CHECKING
 from warnings import warn
 
 import h5py
@@ -16,20 +17,23 @@ from scipy import sparse
 from .. import AnnData
 from ..compat import _deprecate_positional_args
 from .utils import is_float
-from .h5ad import read_h5ad
+
+if TYPE_CHECKING:
+    from collections.abc import Generator, Iterable, Iterator, Mapping
 
 try:
     from .zarr import read_zarr
-except ImportError as e:  # noqa: F841
+except ImportError as _e:
+    e = _e
 
     def read_zarr(*_, **__):
         raise e
 
 
 def read_csv(
-    filename: Union[PathLike, Iterator[str]],
-    delimiter: Optional[str] = ",",
-    first_column_names: Optional[bool] = None,
+    filename: PathLike | Iterator[str],
+    delimiter: str | None = ",",
+    first_column_names: bool | None = None,
     dtype: str = "float32",
 ) -> AnnData:
     """\
@@ -53,9 +57,7 @@ def read_csv(
     return read_text(filename, delimiter, first_column_names, dtype)
 
 
-def read_excel(
-    filename: PathLike, sheet: Union[str, int], dtype: str = "float32"
-) -> AnnData:
+def read_excel(filename: PathLike, sheet: str | int, dtype: str = "float32") -> AnnData:
     """\
     Read `.xlsx` (Excel) file.
 
@@ -76,7 +78,7 @@ def read_excel(
     X = df.values[:, 1:]
     row = dict(row_names=df.iloc[:, 0].values.astype(str))
     col = dict(col_names=np.array(df.columns[1:], dtype=str))
-    return AnnData(X, row, col, dtype=dtype)
+    return AnnData(X, row, col)
 
 
 def read_umi_tools(filename: PathLike, dtype=None) -> AnnData:
@@ -93,15 +95,13 @@ def read_umi_tools(filename: PathLike, dtype=None) -> AnnData:
     table = pd.read_table(filename, dtype={"gene": "category", "cell": "category"})
 
     X = sparse.csr_matrix(
-        (table["count"], (table["cell"].cat.codes, table["gene"].cat.codes))
+        (table["count"], (table["cell"].cat.codes, table["gene"].cat.codes)),
+        dtype=dtype,
     )
     obs = pd.DataFrame(index=pd.Index(table["cell"].cat.categories, name="cell"))
     var = pd.DataFrame(index=pd.Index(table["gene"].cat.categories, name="gene"))
 
-    if dtype is None:
-        dtype = X.dtype
-
-    return AnnData(X=X, obs=obs, var=var, dtype=dtype)
+    return AnnData(X=X, obs=obs, var=var)
 
 
 def read_hdf(filename: PathLike, key: str) -> AnnData:
@@ -133,13 +133,13 @@ def read_hdf(filename: PathLike, key: str) -> AnnData:
         for iname, name in enumerate(["row_names", "col_names"]):
             if name in keys:
                 rows_cols[iname][name] = f[name][()]
-    adata = AnnData(X, rows_cols[0], rows_cols[1], dtype=X.dtype.name)
+    adata = AnnData(X, rows_cols[0], rows_cols[1])
     return adata
 
 
 def _fmt_loom_axis_attrs(
     input: Mapping, idx_name: str, dimm_mapping: Mapping[str, Iterable[str]]
-) -> Tuple[pd.DataFrame, Mapping[str, np.ndarray]]:
+) -> tuple[pd.DataFrame, Mapping[str, np.ndarray]]:
     axis_df = pd.DataFrame()
     axis_mapping = {}
     for key, names in dimm_mapping.items():
@@ -165,9 +165,9 @@ def read_loom(
     cleanup: bool = False,
     X_name: str = "spliced",
     obs_names: str = "CellID",
-    obsm_names: Optional[Mapping[str, Iterable[str]]] = None,
+    obsm_names: Mapping[str, Iterable[str]] | None = None,
     var_names: str = "Gene",
-    varm_names: Optional[Mapping[str, Iterable[str]]] = None,
+    varm_names: Mapping[str, Iterable[str]] | None = None,
     dtype: str = "float32",
     obsm_mapping: Mapping[str, Iterable[str]] = MappingProxyType({}),
     varm_mapping: Mapping[str, Iterable[str]] = MappingProxyType({}),
@@ -228,7 +228,7 @@ def read_loom(
         )
         if obsm_mapping != {}:
             raise ValueError(
-                "Recieved values for both `obsm_names` and `obsm_mapping`. This is "
+                "Received values for both `obsm_names` and `obsm_mapping`. This is "
                 "ambiguous, only pass `obsm_mapping`."
             )
         obsm_mapping = obsm_names
@@ -240,7 +240,7 @@ def read_loom(
         )
         if varm_mapping != {}:
             raise ValueError(
-                "Recieved values for both `varm_names` and `varm_mapping`. This is "
+                "Received values for both `varm_names` and `varm_mapping`. This is "
                 "ambiguous, only pass `varm_mapping`."
             )
         varm_mapping = varm_names
@@ -252,6 +252,7 @@ def read_loom(
         if X_name not in lc.layers.keys():
             X_name = ""
         X = lc.layers[X_name].sparse().T.tocsr() if sparse else lc.layers[X_name][()].T
+        X = X.astype(dtype, copy=False)
 
         layers = OrderedDict()
         if X_name != "":
@@ -273,16 +274,16 @@ def read_loom(
         uns = {}
         if cleanup:
             uns_obs = {}
-            for key in list(obs.keys()):
-                if len(set(obs[key])) == 1:
-                    uns_obs[f"{key}"] = obs[key][0]
+            for key in obs.columns:
+                if len(obs[key].unique()) == 1:
+                    uns_obs[key] = obs[key].iloc[0]
                     del obs[key]
             if uns_obs:
                 uns["loom-obs"] = uns_obs
             uns_var = {}
-            for key in list(var.keys()):
-                if len(set(var[key])) == 1:
-                    uns_var[f"{key}"] = var[key][0]
+            for key in var.columns:
+                if len(var[key].unique()) == 1:
+                    uns_var[key] = var[key].iloc[0]
                     del var[key]
             if uns_var:
                 uns["loom-var"] = uns_var
@@ -295,7 +296,6 @@ def read_loom(
             obsm=obsm if obsm else None,
             varm=varm if varm else None,
             uns=uns,
-            dtype=dtype,
         )
     return adata
 
@@ -318,13 +318,13 @@ def read_mtx(filename: PathLike, dtype: str = "float32") -> AnnData:
     from scipy.sparse import csr_matrix
 
     X = csr_matrix(X)
-    return AnnData(X, dtype=dtype)
+    return AnnData(X)
 
 
 def read_text(
-    filename: Union[PathLike, Iterator[str]],
-    delimiter: Optional[str] = None,
-    first_column_names: Optional[bool] = None,
+    filename: PathLike | Iterator[str],
+    delimiter: str | None = None,
+    first_column_names: bool | None = None,
     dtype: str = "float32",
 ) -> AnnData:
     """\
@@ -360,7 +360,7 @@ def read_text(
             return _read_text(f, delimiter, first_column_names, dtype)
 
 
-def iter_lines(file_like: Iterable[str]) -> Generator[str, None, None]:
+def _iter_lines(file_like: Iterable[str]) -> Generator[str, None, None]:
     """Helper for iterating only nonempty lines without line breaks"""
     for line in file_like:
         line = line.rstrip("\r\n")
@@ -370,13 +370,13 @@ def iter_lines(file_like: Iterable[str]) -> Generator[str, None, None]:
 
 def _read_text(
     f: Iterator[str],
-    delimiter: Optional[str],
-    first_column_names: Optional[bool],
+    delimiter: str | None,
+    first_column_names: bool | None,
     dtype: str,
 ) -> AnnData:
     comments = []
     data = []
-    lines = iter_lines(f)
+    lines = _iter_lines(f)
     col_names = []
     row_names = []
     # read header and column names
@@ -443,7 +443,7 @@ def _read_text(
         else:
             data.append(np.array(line_list, dtype=dtype))
     # logg.msg("    read data into list of lists", t=True, v=4)
-    # transfrom to array, this takes a long time and a lot of memory
+    # transform to array, this takes a long time and a lot of memory
     # but it’s actually the same thing as np.genfromtxt does
     # - we don’t use the latter as it would involve another slicing step
     #   in the end, to separate row_names from float data, slicing takes
@@ -472,24 +472,4 @@ def _read_text(
         data,
         obs=dict(obs_names=row_names),
         var=dict(var_names=col_names),
-        dtype=dtype,
     )
-
-
-def load_sparse_csr(d, key="X"):
-    from scipy.sparse.csr import csr_matrix
-
-    key_csr = f"{key}_csr"
-    d[key] = csr_matrix(
-        (d[f"{key_csr}_data"], d[f"{key_csr}_indices"], d[f"{key_csr}_indptr"]),
-        shape=d[f"{key_csr}_shape"],
-    )
-    del_sparse_matrix_keys(d, key_csr)
-    return d
-
-
-def del_sparse_matrix_keys(mapping, key_csr):
-    del mapping[f"{key_csr}_data"]
-    del mapping[f"{key_csr}_indices"]
-    del mapping[f"{key_csr}_indptr"]
-    del mapping[f"{key_csr}_shape"]
