@@ -1,10 +1,10 @@
 from __future__ import annotations
 
-from os import PathLike
 from collections.abc import Mapping
 from functools import partial
-from typing import Union
+from itertools import product
 from types import MappingProxyType
+from typing import TYPE_CHECKING, Literal
 from warnings import warn
 
 import h5py
@@ -14,35 +14,47 @@ from scipy import sparse
 
 import anndata as ad
 from anndata import AnnData, Raw
+from anndata._core import views
 from anndata._core.index import _normalize_indices
 from anndata._core.merge import intersect_keys
-from anndata._core.sparse_dataset import SparseDataset
-from anndata._core import views
+from anndata._core.sparse_dataset import CSCDataset, CSRDataset, sparse_dataset
+from anndata._io.utils import H5PY_V3, check_key
+from anndata._warnings import OldFormatWarning
 from anndata.compat import (
-    Literal,
-    OverloadedDict,
+    AwkArray,
+    CupyArray,
+    CupyCSCMatrix,
+    CupyCSRMatrix,
+    DaskArray,
     ZarrArray,
     ZarrGroup,
-    _read_attr,
-    _from_fixed_length_strings,
     _decode_structured_array,
+    _from_fixed_length_strings,
+    _read_attr,
 )
-from anndata._io.utils import report_write_key_on_error, check_key, H5PY_V3
-from anndata._warnings import OldFormatWarning
 
-from .registry import (
-    _REGISTRY,
-    IOSpec,
-    get_spec,
-    read_elem,
-    read_elem_partial,
-    write_elem,
-)
+from .registry import _REGISTRY, IOSpec, read_elem, read_elem_partial
+
+if TYPE_CHECKING:
+    from os import PathLike
 
 H5Array = h5py.Dataset
 H5Group = h5py.Group
+H5File = h5py.File
 
 
+####################
+# Dask utils       #
+####################
+
+try:
+    from dask.utils import SerializableLock as Lock
+except ImportError:
+    from threading import Lock
+
+# to fix https://github.com/dask/distributed/issues/780
+GLOBAL_LOCK = Lock()
+
 ####################
 # Dispatch methods #
 ####################
@@ -67,6 +79,26 @@ H5Group = h5py.Group
 #     return False
 
 
+def _to_cpu_mem_wrapper(write_func):
+    """
+    Wrapper to bring cupy types into cpu memory before writing.
+
+    Ideally we do direct writing at some point.
+    """
+
+    def wrapper(
+        f,
+        k,
+        cupy_val: CupyArray | CupyCSCMatrix | CupyCSRMatrix,
+        _writer,
+        *,
+        dataset_kwargs=MappingProxyType,
+    ):
+        return write_func(f, k, cupy_val.get(), _writer, dataset_kwargs=dataset_kwargs)
+
+    return wrapper
+
+
 ################################
 # Fallbacks / backwards compat #
 ################################
@@ -74,9 +106,10 @@ H5Group = h5py.Group
 # Note: there is no need for writing in a backwards compatible format, maybe
 
 
+@_REGISTRY.register_read(H5File, IOSpec("", ""))
 @_REGISTRY.register_read(H5Group, IOSpec("", ""))
 @_REGISTRY.register_read(H5Array, IOSpec("", ""))
-def read_basic(elem):
+def read_basic(elem, _reader):
     from anndata._io import h5ad
 
     warn(
@@ -88,15 +121,15 @@ def read_basic(elem):
     if isinstance(elem, Mapping):
         # Backwards compat sparse arrays
         if "h5sparse_format" in elem.attrs:
-            return SparseDataset(elem).to_memory()
-        return {k: read_elem(v) for k, v in elem.items()}
+            return sparse_dataset(elem).to_memory()
+        return {k: _reader.read_elem(v) for k, v in elem.items()}
     elif isinstance(elem, h5py.Dataset):
         return h5ad.read_dataset(elem)  # TODO: Handle legacy
 
 
 @_REGISTRY.register_read(ZarrGroup, IOSpec("", ""))
 @_REGISTRY.register_read(ZarrArray, IOSpec("", ""))
-def read_basic_zarr(elem):
+def read_basic_zarr(elem, _reader):
     from anndata._io import zarr
 
     warn(
@@ -108,8 +141,8 @@ def read_basic_zarr(elem):
     if isinstance(elem, Mapping):
         # Backwards compat sparse arrays
         if "h5sparse_format" in elem.attrs:
-            return SparseDataset(elem).to_memory()
-        return {k: read_elem(v) for k, v in elem.items()}
+            return sparse_dataset(elem).to_memory()
+        return {k: _reader.read_elem(v) for k, v in elem.items()}
     elif isinstance(elem, ZarrArray):
         return zarr.read_dataset(elem)  # TODO: Handle legacy
 
@@ -212,25 +245,27 @@ def _read_partial(group, *, items=None, indices=(slice(None), slice(None))):
 
 @_REGISTRY.register_write(ZarrGroup, AnnData, IOSpec("anndata", "0.1.0"))
 @_REGISTRY.register_write(H5Group, AnnData, IOSpec("anndata", "0.1.0"))
-def write_anndata(f, k, adata, dataset_kwargs=MappingProxyType({})):
+def write_anndata(f, k, adata, _writer, dataset_kwargs=MappingProxyType({})):
     g = f.require_group(k)
-    write_elem(g, "X", adata.X, dataset_kwargs=dataset_kwargs)
-    write_elem(g, "obs", adata.obs, dataset_kwargs=dataset_kwargs)
-    write_elem(g, "var", adata.var, dataset_kwargs=dataset_kwargs)
-    write_elem(g, "obsm", dict(adata.obsm), dataset_kwargs=dataset_kwargs)
-    write_elem(g, "varm", dict(adata.varm), dataset_kwargs=dataset_kwargs)
-    write_elem(g, "obsp", dict(adata.obsp), dataset_kwargs=dataset_kwargs)
-    write_elem(g, "varp", dict(adata.varp), dataset_kwargs=dataset_kwargs)
-    write_elem(g, "layers", dict(adata.layers), dataset_kwargs=dataset_kwargs)
-    write_elem(g, "uns", dict(adata.uns), dataset_kwargs=dataset_kwargs)
-    write_elem(g, "raw", adata.raw, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "X", adata.X, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "obs", adata.obs, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "var", adata.var, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "obsm", dict(adata.obsm), dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "varm", dict(adata.varm), dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "obsp", dict(adata.obsp), dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "varp", dict(adata.varp), dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "layers", dict(adata.layers), dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "uns", dict(adata.uns), dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "raw", adata.raw, dataset_kwargs=dataset_kwargs)
 
 
 @_REGISTRY.register_read(H5Group, IOSpec("anndata", "0.1.0"))
 @_REGISTRY.register_read(H5Group, IOSpec("raw", "0.1.0"))
+@_REGISTRY.register_read(H5File, IOSpec("anndata", "0.1.0"))
+@_REGISTRY.register_read(H5File, IOSpec("raw", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("anndata", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("raw", "0.1.0"))
-def read_anndata(elem):
+def read_anndata(elem, _reader):
     d = {}
     for k in [
         "X",
@@ -245,19 +280,17 @@ def read_anndata(elem):
         "raw",
     ]:
         if k in elem:
-            d[k] = read_elem(elem[k])
-        if "X" in d:
-            d["dtype"] = d["X"].dtype
+            d[k] = _reader.read_elem(elem[k])
     return AnnData(**d)
 
 
 @_REGISTRY.register_write(H5Group, Raw, IOSpec("raw", "0.1.0"))
 @_REGISTRY.register_write(ZarrGroup, Raw, IOSpec("raw", "0.1.0"))
-def write_raw(f, k, raw, dataset_kwargs=MappingProxyType({})):
-    g = f.create_group(k)
-    write_elem(g, "X", raw.X, dataset_kwargs=dataset_kwargs)
-    write_elem(g, "var", raw.var, dataset_kwargs=dataset_kwargs)
-    write_elem(g, "varm", dict(raw.varm), dataset_kwargs=dataset_kwargs)
+def write_raw(f, k, raw, _writer, dataset_kwargs=MappingProxyType({})):
+    g = f.require_group(k)
+    _writer.write_elem(g, "X", raw.X, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "var", raw.var, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "varm", dict(raw.varm), dataset_kwargs=dataset_kwargs)
 
 
 ############
@@ -267,18 +300,16 @@ def write_raw(f, k, raw, dataset_kwargs=MappingProxyType({})):
 
 @_REGISTRY.register_read(H5Group, IOSpec("dict", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("dict", "0.1.0"))
-def read_mapping(elem):
-    return {k: read_elem(v) for k, v in elem.items()}
+def read_mapping(elem, _reader):
+    return {k: _reader.read_elem(v) for k, v in elem.items()}
 
 
-@_REGISTRY.register_write(H5Group, OverloadedDict, IOSpec("dict", "0.1.0"))
 @_REGISTRY.register_write(H5Group, dict, IOSpec("dict", "0.1.0"))
-@_REGISTRY.register_write(ZarrGroup, OverloadedDict, IOSpec("dict", "0.1.0"))
 @_REGISTRY.register_write(ZarrGroup, dict, IOSpec("dict", "0.1.0"))
-def write_mapping(f, k, v, dataset_kwargs=MappingProxyType({})):
-    g = f.create_group(k)
+def write_mapping(f, k, v, _writer, dataset_kwargs=MappingProxyType({})):
+    g = f.require_group(k)
     for sub_k, sub_v in v.items():
-        write_elem(g, sub_k, sub_v, dataset_kwargs=dataset_kwargs)
+        _writer.write_elem(g, sub_k, sub_v, dataset_kwargs=dataset_kwargs)
 
 
 ##############
@@ -288,11 +319,11 @@ def write_mapping(f, k, v, dataset_kwargs=MappingProxyType({})):
 
 @_REGISTRY.register_write(H5Group, list, IOSpec("array", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, list, IOSpec("array", "0.2.0"))
-def write_list(f, k, elem, dataset_kwargs=MappingProxyType({})):
-    write_elem(f, k, np.array(elem), dataset_kwargs=dataset_kwargs)
+def write_list(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
+    _writer.write_elem(f, k, np.array(elem), dataset_kwargs=dataset_kwargs)
 
 
-# TODO: Is this the right behaviour for MaskedArrays?
+# TODO: Is this the right behavior for MaskedArrays?
 # It's in the `AnnData.concatenate` docstring, but should we keep it?
 @_REGISTRY.register_write(H5Group, views.ArrayView, IOSpec("array", "0.2.0"))
 @_REGISTRY.register_write(H5Group, np.ndarray, IOSpec("array", "0.2.0"))
@@ -302,33 +333,70 @@ def write_list(f, k, elem, dataset_kwargs=MappingProxyType({})):
 @_REGISTRY.register_write(ZarrGroup, np.ndarray, IOSpec("array", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, h5py.Dataset, IOSpec("array", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, np.ma.MaskedArray, IOSpec("array", "0.2.0"))
-def write_basic(f, k, elem, dataset_kwargs=MappingProxyType({})):
-    """Write methods which underlying library handles nativley."""
+@_REGISTRY.register_write(ZarrGroup, ZarrArray, IOSpec("array", "0.2.0"))
+def write_basic(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
+    """Write methods which underlying library handles natively."""
     f.create_dataset(k, data=elem, **dataset_kwargs)
 
 
+_REGISTRY.register_write(H5Group, CupyArray, IOSpec("array", "0.2.0"))(
+    _to_cpu_mem_wrapper(write_basic)
+)
+_REGISTRY.register_write(ZarrGroup, CupyArray, IOSpec("array", "0.2.0"))(
+    _to_cpu_mem_wrapper(write_basic)
+)
+
+
+@_REGISTRY.register_write(ZarrGroup, DaskArray, IOSpec("array", "0.2.0"))
+def write_basic_dask_zarr(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
+    import dask.array as da
+
+    g = f.require_dataset(k, shape=elem.shape, dtype=elem.dtype, **dataset_kwargs)
+    da.store(elem, g, lock=GLOBAL_LOCK)
+
+
+# Adding this separately because h5py isn't serializable
+# https://github.com/pydata/xarray/issues/4242
+@_REGISTRY.register_write(H5Group, DaskArray, IOSpec("array", "0.2.0"))
+def write_basic_dask_h5(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
+    import dask.array as da
+    import dask.config as dc
+
+    if dc.get("scheduler", None) == "dask.distributed":
+        raise ValueError(
+            "Cannot write dask arrays to hdf5 when using distributed scheduler"
+        )
+
+    g = f.require_dataset(k, shape=elem.shape, dtype=elem.dtype, **dataset_kwargs)
+    da.store(elem, g)
+
+
 @_REGISTRY.register_read(H5Array, IOSpec("array", "0.2.0"))
 @_REGISTRY.register_read(ZarrArray, IOSpec("array", "0.2.0"))
 @_REGISTRY.register_read(ZarrArray, IOSpec("string-array", "0.2.0"))
-def read_array(elem):
+def read_array(elem, _reader):
     return elem[()]
 
 
 @_REGISTRY.register_read_partial(H5Array, IOSpec("array", "0.2.0"))
-@_REGISTRY.register_read_partial(ZarrArray, IOSpec("array", "0.2.0"))
 @_REGISTRY.register_read_partial(ZarrArray, IOSpec("string-array", "0.2.0"))
 def read_array_partial(elem, *, items=None, indices=(slice(None, None))):
     return elem[indices]
 
 
+@_REGISTRY.register_read_partial(ZarrArray, IOSpec("array", "0.2.0"))
+def read_zarr_array_partial(elem, *, items=None, indices=(slice(None, None))):
+    return elem.oindex[indices]
+
+
 # arrays of strings
 @_REGISTRY.register_read(H5Array, IOSpec("string-array", "0.2.0"))
-def read_string_array(d):
-    return read_array(d.asstr())
+def read_string_array(d, _reader):
+    return read_array(d.asstr(), _reader=_reader)
 
 
 @_REGISTRY.register_read_partial(H5Array, IOSpec("string-array", "0.2.0"))
-def read_array_partial(d, items=None, indices=slice(None)):
+def read_string_array_partial(d, items=None, indices=slice(None)):
     return read_array_partial(d.asstr(), items=items, indices=indices)
 
 
@@ -340,7 +408,7 @@ def read_array_partial(d, items=None, indices=slice(None)):
 )
 @_REGISTRY.register_write(H5Group, (np.ndarray, "U"), IOSpec("string-array", "0.2.0"))
 @_REGISTRY.register_write(H5Group, (np.ndarray, "O"), IOSpec("string-array", "0.2.0"))
-def write_vlen_string_array(f, k, elem, dataset_kwargs=MappingProxyType({})):
+def write_vlen_string_array(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
     """Write methods which underlying library handles nativley."""
     str_dtype = h5py.special_dtype(vlen=str)
     f.create_dataset(k, data=elem.astype(str_dtype), dtype=str_dtype, **dataset_kwargs)
@@ -354,7 +422,9 @@ def write_vlen_string_array(f, k, elem, dataset_kwargs=MappingProxyType({})):
 )
 @_REGISTRY.register_write(ZarrGroup, (np.ndarray, "U"), IOSpec("string-array", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, (np.ndarray, "O"), IOSpec("string-array", "0.2.0"))
-def write_vlen_string_array_zarr(f, k, elem, dataset_kwargs=MappingProxyType({})):
+def write_vlen_string_array_zarr(
+    f, k, elem, _writer, dataset_kwargs=MappingProxyType({})
+):
     import numcodecs
 
     f.create_dataset(
@@ -385,7 +455,7 @@ def _to_hdf5_vlen_strings(value: np.ndarray) -> np.ndarray:
 
 @_REGISTRY.register_read(H5Array, IOSpec("rec-array", "0.2.0"))
 @_REGISTRY.register_read(ZarrArray, IOSpec("rec-array", "0.2.0"))
-def read_recarray(d):
+def read_recarray(d, _reader):
     value = d[()]
     dtype = value.dtype
     value = _from_fixed_length_strings(value)
@@ -396,13 +466,13 @@ def read_recarray(d):
 
 @_REGISTRY.register_write(H5Group, (np.ndarray, "V"), IOSpec("rec-array", "0.2.0"))
 @_REGISTRY.register_write(H5Group, np.recarray, IOSpec("rec-array", "0.2.0"))
-def write_recarray(f, k, elem, dataset_kwargs=MappingProxyType({})):
+def write_recarray(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
     f.create_dataset(k, data=_to_hdf5_vlen_strings(elem), **dataset_kwargs)
 
 
 @_REGISTRY.register_write(ZarrGroup, (np.ndarray, "V"), IOSpec("rec-array", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, np.recarray, IOSpec("rec-array", "0.2.0"))
-def write_recarray_zarr(f, k, elem, dataset_kwargs=MappingProxyType({})):
+def write_recarray_zarr(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
     from anndata.compat import _to_fixed_length_strings
 
     f.create_dataset(k, data=_to_fixed_length_strings(elem), **dataset_kwargs)
@@ -414,13 +484,18 @@ def write_recarray_zarr(f, k, elem, dataset_kwargs=MappingProxyType({})):
 
 
 def write_sparse_compressed(
-    f, key, value, fmt: Literal["csr", "csc"], dataset_kwargs=MappingProxyType({})
+    f,
+    key,
+    value,
+    _writer,
+    fmt: Literal["csr", "csc"],
+    dataset_kwargs=MappingProxyType({}),
 ):
-    g = f.create_group(key)
+    g = f.require_group(key)
     g.attrs["shape"] = value.shape
 
-    # Allow resizing
-    if "maxshape" not in dataset_kwargs:
+    # Allow resizing for hdf5
+    if isinstance(f, H5Group) and "maxshape" not in dataset_kwargs:
         dataset_kwargs = dict(maxshape=(None,), **dataset_kwargs)
 
     g.create_dataset("data", data=value.data, **dataset_kwargs)
@@ -430,55 +505,153 @@ def write_sparse_compressed(
 
 write_csr = partial(write_sparse_compressed, fmt="csr")
 write_csc = partial(write_sparse_compressed, fmt="csc")
-_REGISTRY.register_write(H5Group, sparse.csr_matrix, IOSpec("csr_matrix", "0.1.0"))(
-    write_csr
-)
-_REGISTRY.register_write(H5Group, views.SparseCSRView, IOSpec("csr_matrix", "0.1.0"))(
-    write_csr
-)
-_REGISTRY.register_write(H5Group, sparse.csc_matrix, IOSpec("csc_matrix", "0.1.0"))(
-    write_csc
-)
-_REGISTRY.register_write(H5Group, views.SparseCSCView, IOSpec("csc_matrix", "0.1.0"))(
-    write_csc
-)
-_REGISTRY.register_write(ZarrGroup, sparse.csr_matrix, IOSpec("csr_matrix", "0.1.0"))(
-    write_csr
-)
-_REGISTRY.register_write(ZarrGroup, views.SparseCSRView, IOSpec("csr_matrix", "0.1.0"))(
-    write_csr
-)
-_REGISTRY.register_write(ZarrGroup, sparse.csc_matrix, IOSpec("csc_matrix", "0.1.0"))(
-    write_csc
-)
-_REGISTRY.register_write(ZarrGroup, views.SparseCSCView, IOSpec("csc_matrix", "0.1.0"))(
-    write_csc
-)
+
+for store_type, (cls, spec, func) in product(
+    (H5Group, ZarrGroup),
+    [
+        (sparse.csr_matrix, IOSpec("csr_matrix", "0.1.0"), write_csr),
+        (views.SparseCSRView, IOSpec("csr_matrix", "0.1.0"), write_csr),
+        (sparse.csc_matrix, IOSpec("csc_matrix", "0.1.0"), write_csc),
+        (views.SparseCSCView, IOSpec("csc_matrix", "0.1.0"), write_csc),
+        (CupyCSRMatrix, IOSpec("csr_matrix", "0.1.0"), _to_cpu_mem_wrapper(write_csr)),
+        (
+            views.CupySparseCSRView,
+            IOSpec("csr_matrix", "0.1.0"),
+            _to_cpu_mem_wrapper(write_csr),
+        ),
+        (CupyCSCMatrix, IOSpec("csc_matrix", "0.1.0"), _to_cpu_mem_wrapper(write_csc)),
+        (
+            views.CupySparseCSCView,
+            IOSpec("csc_matrix", "0.1.0"),
+            _to_cpu_mem_wrapper(write_csc),
+        ),
+    ],
+):
+    _REGISTRY.register_write(store_type, cls, spec)(func)
 
 
-@_REGISTRY.register_write(H5Group, SparseDataset, IOSpec("", "0.1.0"))
-@_REGISTRY.register_write(ZarrGroup, SparseDataset, IOSpec("", "0.1.0"))
-def write_sparse_dataset(f, k, elem, dataset_kwargs=MappingProxyType({})):
+@_REGISTRY.register_write(H5Group, CSRDataset, IOSpec("", "0.1.0"))
+@_REGISTRY.register_write(H5Group, CSCDataset, IOSpec("", "0.1.0"))
+@_REGISTRY.register_write(ZarrGroup, CSRDataset, IOSpec("", "0.1.0"))
+@_REGISTRY.register_write(ZarrGroup, CSCDataset, IOSpec("", "0.1.0"))
+def write_sparse_dataset(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
     write_sparse_compressed(
-        f, k, elem.to_backed(), fmt=elem.format_str, dataset_kwargs=dataset_kwargs
+        f,
+        k,
+        elem._to_backed(),
+        _writer,
+        fmt=elem.format,
+        dataset_kwargs=dataset_kwargs,
     )
     # TODO: Cleaner way to do this
-    f[k].attrs["encoding-type"] = f"{elem.format_str}_matrix"
+    f[k].attrs["encoding-type"] = f"{elem.format}_matrix"
     f[k].attrs["encoding-version"] = "0.1.0"
 
 
+@_REGISTRY.register_write(
+    H5Group, (DaskArray, sparse.csr_matrix), IOSpec("csr_matrix", "0.1.0")
+)
+@_REGISTRY.register_write(
+    H5Group, (DaskArray, sparse.csc_matrix), IOSpec("csc_matrix", "0.1.0")
+)
+@_REGISTRY.register_write(
+    ZarrGroup, (DaskArray, sparse.csr_matrix), IOSpec("csr_matrix", "0.1.0")
+)
+@_REGISTRY.register_write(
+    ZarrGroup, (DaskArray, sparse.csc_matrix), IOSpec("csc_matrix", "0.1.0")
+)
+def write_dask_sparse(f, k, elem, _writer, dataset_kwargs=MappingProxyType({})):
+    sparse_format = elem._meta.format
+
+    def as_int64_indices(x):
+        x.indptr = x.indptr.astype(np.int64, copy=False)
+        x.indices = x.indices.astype(np.int64, copy=False)
+        return x
+
+    if sparse_format == "csr":
+        axis = 0
+    elif sparse_format == "csc":
+        axis = 1
+    else:
+        raise NotImplementedError(
+            f"Cannot write dask sparse arrays with format {sparse_format}"
+        )
+
+    def chunk_slice(start: int, stop: int) -> tuple[slice | None, slice | None]:
+        result = [slice(None), slice(None)]
+        result[axis] = slice(start, stop)
+        return tuple(result)
+
+    axis_chunks = elem.chunks[axis]
+    chunk_start = 0
+    chunk_stop = axis_chunks[0]
+
+    _writer.write_elem(
+        f,
+        k,
+        as_int64_indices(elem[chunk_slice(chunk_start, chunk_stop)].compute()),
+        dataset_kwargs=dataset_kwargs,
+    )
+
+    disk_mtx = sparse_dataset(f[k])
+
+    for chunk_size in axis_chunks[1:]:
+        chunk_start = chunk_stop
+        chunk_stop += chunk_size
+
+        disk_mtx.append(elem[chunk_slice(chunk_start, chunk_stop)].compute())
+
+
 @_REGISTRY.register_read(H5Group, IOSpec("csc_matrix", "0.1.0"))
 @_REGISTRY.register_read(H5Group, IOSpec("csr_matrix", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("csc_matrix", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("csr_matrix", "0.1.0"))
-def read_sparse(elem):
-    return SparseDataset(elem).to_memory()
+def read_sparse(elem, _reader):
+    return sparse_dataset(elem).to_memory()
 
 
 @_REGISTRY.register_read_partial(H5Group, IOSpec("csc_matrix", "0.1.0"))
 @_REGISTRY.register_read_partial(H5Group, IOSpec("csr_matrix", "0.1.0"))
+@_REGISTRY.register_read_partial(ZarrGroup, IOSpec("csc_matrix", "0.1.0"))
+@_REGISTRY.register_read_partial(ZarrGroup, IOSpec("csr_matrix", "0.1.0"))
 def read_sparse_partial(elem, *, items=None, indices=(slice(None), slice(None))):
-    return SparseDataset(elem)[indices]
+    return sparse_dataset(elem)[indices]
+
+
+#################
+# Awkward array #
+#################
+
+
+@_REGISTRY.register_write(H5Group, AwkArray, IOSpec("awkward-array", "0.1.0"))
+@_REGISTRY.register_write(ZarrGroup, AwkArray, IOSpec("awkward-array", "0.1.0"))
+@_REGISTRY.register_write(
+    H5Group, views.AwkwardArrayView, IOSpec("awkward-array", "0.1.0")
+)
+@_REGISTRY.register_write(
+    ZarrGroup, views.AwkwardArrayView, IOSpec("awkward-array", "0.1.0")
+)
+def write_awkward(f, k, v, _writer, dataset_kwargs=MappingProxyType({})):
+    from anndata.compat import awkward as ak
+
+    group = f.require_group(k)
+    form, length, container = ak.to_buffers(ak.to_packed(v))
+    group.attrs["length"] = length
+    group.attrs["form"] = form.to_json()
+    for k, v in container.items():
+        _writer.write_elem(group, k, v, dataset_kwargs=dataset_kwargs)
+
+
+@_REGISTRY.register_read(H5Group, IOSpec("awkward-array", "0.1.0"))
+@_REGISTRY.register_read(ZarrGroup, IOSpec("awkward-array", "0.1.0"))
+def read_awkward(elem, _reader):
+    from anndata.compat import awkward as ak
+
+    form = _read_attr(elem.attrs, "form")
+    length = _read_attr(elem.attrs, "length")
+    container = {k: _reader.read_elem(elem[k]) for k in elem.keys()}
+
+    return ak.from_buffers(form, int(length), container)
 
 
 ##############
@@ -490,16 +663,29 @@ def read_sparse_partial(elem, *, items=None, indices=(slice(None), slice(None)))
 @_REGISTRY.register_write(H5Group, pd.DataFrame, IOSpec("dataframe", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, views.DataFrameView, IOSpec("dataframe", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, pd.DataFrame, IOSpec("dataframe", "0.2.0"))
-def write_dataframe(f, key, df, dataset_kwargs=MappingProxyType({})):
+def write_dataframe(f, key, df, _writer, dataset_kwargs=MappingProxyType({})):
     # Check arguments
     for reserved in ("_index",):
         if reserved in df.columns:
             raise ValueError(f"{reserved!r} is a reserved name for dataframe columns.")
-    group = f.create_group(key)
+    group = f.require_group(key)
+    if not df.columns.is_unique:
+        duplicates = list(df.columns[df.columns.duplicated()])
+        raise ValueError(
+            f"Found repeated column names: {duplicates}. Column names must be unique."
+        )
     col_names = [check_key(c) for c in df.columns]
     group.attrs["column-order"] = col_names
 
     if df.index.name is not None:
+        if df.index.name in col_names and not pd.Series(
+            df.index, index=df.index
+        ).equals(df[df.index.name]):
+            raise ValueError(
+                f"DataFrame.index.name ({df.index.name!r}) is also used by a column "
+                "whose values are different. This is not supported. Please make sure "
+                "the values are the same, or use a different name."
+            )
         index_name = df.index.name
     else:
         index_name = "_index"
@@ -508,21 +694,25 @@ def write_dataframe(f, key, df, dataset_kwargs=MappingProxyType({})):
     # ._values is "the best" array representation. It's the true array backing the
     # object, where `.values` is always a np.ndarray and .array is always a pandas
     # array.
-    write_elem(group, index_name, df.index._values, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(
+        group, index_name, df.index._values, dataset_kwargs=dataset_kwargs
+    )
     for colname, series in df.items():
         # TODO: this should write the "true" representation of the series (i.e. the underlying array or ndarray depending)
-        write_elem(group, colname, series._values, dataset_kwargs=dataset_kwargs)
+        _writer.write_elem(
+            group, colname, series._values, dataset_kwargs=dataset_kwargs
+        )
 
 
 @_REGISTRY.register_read(H5Group, IOSpec("dataframe", "0.2.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("dataframe", "0.2.0"))
-def read_dataframe(elem):
+def read_dataframe(elem, _reader):
     columns = list(_read_attr(elem.attrs, "column-order"))
     idx_key = _read_attr(elem.attrs, "_index")
     df = pd.DataFrame(
-        {k: read_elem(elem[k]) for k in columns},
-        index=read_elem(elem[idx_key]),
-        columns=list(columns),
+        {k: _reader.read_elem(elem[k]) for k in columns},
+        index=_reader.read_elem(elem[idx_key]),
+        columns=columns if len(columns) else None,
     )
     if idx_key != "_index":
         df.index.name = idx_key
@@ -545,7 +735,7 @@ def read_dataframe_partial(
     df = pd.DataFrame(
         {k: read_elem_partial(elem[k], indices=indices[0]) for k in columns},
         index=read_elem_partial(elem[idx_key], indices=indices[0]),
-        columns=list(columns),
+        columns=columns if len(columns) else None,
     )
     if idx_key != "_index":
         df.index.name = idx_key
@@ -557,20 +747,20 @@ def read_dataframe_partial(
 
 @_REGISTRY.register_read(H5Group, IOSpec("dataframe", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("dataframe", "0.1.0"))
-def read_dataframe_0_1_0(elem):
+def read_dataframe_0_1_0(elem, _reader):
     columns = _read_attr(elem.attrs, "column-order")
     idx_key = _read_attr(elem.attrs, "_index")
     df = pd.DataFrame(
         {k: read_series(elem[k]) for k in columns},
         index=read_series(elem[idx_key]),
-        columns=list(columns),
+        columns=columns if len(columns) else None,
     )
     if idx_key != "_index":
         df.index.name = idx_key
     return df
 
 
-def read_series(dataset: h5py.Dataset) -> Union[np.ndarray, pd.Categorical]:
+def read_series(dataset: h5py.Dataset) -> np.ndarray | pd.Categorical:
     # For reading older dataframes
     if "categories" in dataset.attrs:
         if isinstance(dataset, ZarrArray):
@@ -609,31 +799,33 @@ def read_partial_dataframe_0_1_0(
 
 @_REGISTRY.register_write(H5Group, pd.Categorical, IOSpec("categorical", "0.2.0"))
 @_REGISTRY.register_write(ZarrGroup, pd.Categorical, IOSpec("categorical", "0.2.0"))
-def write_categorical(f, k, v, dataset_kwargs=MappingProxyType({})):
-    g = f.create_group(k)
+def write_categorical(f, k, v, _writer, dataset_kwargs=MappingProxyType({})):
+    g = f.require_group(k)
     g.attrs["ordered"] = bool(v.ordered)
 
-    write_elem(g, "codes", v.codes, dataset_kwargs=dataset_kwargs)
-    write_elem(g, "categories", v.categories._values, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "codes", v.codes, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(
+        g, "categories", v.categories._values, dataset_kwargs=dataset_kwargs
+    )
 
 
 @_REGISTRY.register_read(H5Group, IOSpec("categorical", "0.2.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("categorical", "0.2.0"))
-def read_categorical(elem):
+def read_categorical(elem, _reader):
     return pd.Categorical.from_codes(
-        codes=read_elem(elem["codes"]),
-        categories=read_elem(elem["categories"]),
-        ordered=_read_attr(elem.attrs, "ordered"),
+        codes=_reader.read_elem(elem["codes"]),
+        categories=_reader.read_elem(elem["categories"]),
+        ordered=bool(_read_attr(elem.attrs, "ordered")),
     )
 
 
 @_REGISTRY.register_read_partial(H5Group, IOSpec("categorical", "0.2.0"))
 @_REGISTRY.register_read_partial(ZarrGroup, IOSpec("categorical", "0.2.0"))
-def read_categorical(elem, *, items=None, indices=(slice(None),)):
+def read_partial_categorical(elem, *, items=None, indices=(slice(None),)):
     return pd.Categorical.from_codes(
         codes=read_elem_partial(elem["codes"], indices=indices),
         categories=read_elem(elem["categories"]),
-        ordered=_read_attr(elem.attrs, "ordered"),
+        ordered=bool(_read_attr(elem.attrs, "ordered")),
     )
 
 
@@ -654,33 +846,33 @@ def read_categorical(elem, *, items=None, indices=(slice(None),)):
 @_REGISTRY.register_write(
     ZarrGroup, pd.arrays.BooleanArray, IOSpec("nullable-boolean", "0.1.0")
 )
-def write_nullable_integer(f, k, v, dataset_kwargs=MappingProxyType({})):
-    g = f.create_group(k)
+def write_nullable_integer(f, k, v, _writer, dataset_kwargs=MappingProxyType({})):
+    g = f.require_group(k)
     if v._mask is not None:
-        write_elem(g, "mask", v._mask, dataset_kwargs=dataset_kwargs)
-    write_elem(g, "values", v._data, dataset_kwargs=dataset_kwargs)
+        _writer.write_elem(g, "mask", v._mask, dataset_kwargs=dataset_kwargs)
+    _writer.write_elem(g, "values", v._data, dataset_kwargs=dataset_kwargs)
 
 
 @_REGISTRY.register_read(H5Group, IOSpec("nullable-integer", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("nullable-integer", "0.1.0"))
-def read_nullable_integer(elem):
+def read_nullable_integer(elem, _reader):
     if "mask" in elem:
         return pd.arrays.IntegerArray(
-            read_elem(elem["values"]), mask=read_elem(elem["mask"])
+            _reader.read_elem(elem["values"]), mask=_reader.read_elem(elem["mask"])
         )
     else:
-        return pd.array(read_elem(elem["values"]))
+        return pd.array(_reader.read_elem(elem["values"]))
 
 
 @_REGISTRY.register_read(H5Group, IOSpec("nullable-boolean", "0.1.0"))
 @_REGISTRY.register_read(ZarrGroup, IOSpec("nullable-boolean", "0.1.0"))
-def read_nullable_boolean(elem):
+def read_nullable_boolean(elem, _reader):
     if "mask" in elem:
         return pd.arrays.BooleanArray(
-            read_elem(elem["values"]), mask=read_elem(elem["mask"])
+            _reader.read_elem(elem["values"]), mask=_reader.read_elem(elem["mask"])
         )
     else:
-        return pd.array(read_elem(elem["values"]))
+        return pd.array(_reader.read_elem(elem["values"]))
 
 
 ###########
@@ -690,15 +882,15 @@ def read_nullable_boolean(elem):
 
 @_REGISTRY.register_read(H5Array, IOSpec("numeric-scalar", "0.2.0"))
 @_REGISTRY.register_read(ZarrArray, IOSpec("numeric-scalar", "0.2.0"))
-def read_scalar(elem):
+def read_scalar(elem, _reader):
     return elem[()]
 
 
-def write_scalar(f, key, value, dataset_kwargs=MappingProxyType({})):
+def write_scalar(f, key, value, _writer, dataset_kwargs=MappingProxyType({})):
     return f.create_dataset(key, data=np.array(value), **dataset_kwargs)
 
 
-def write_hdf5_scalar(f, key, value, dataset_kwargs=MappingProxyType({})):
+def write_hdf5_scalar(f, key, value, _writer, dataset_kwargs=MappingProxyType({})):
     # Can’t compress scalars, error is thrown
     dataset_kwargs = dataset_kwargs.copy()
     dataset_kwargs.pop("compression", None)
@@ -723,12 +915,12 @@ _REGISTRY.register_write(ZarrGroup, np.str_, IOSpec("string", "0.2.0"))(write_sc
 
 
 @_REGISTRY.register_read(H5Array, IOSpec("string", "0.2.0"))
-def read_hdf5_string(elem):
+def read_hdf5_string(elem, _reader):
     return elem.asstr()[()]
 
 
 @_REGISTRY.register_read(ZarrArray, IOSpec("string", "0.2.0"))
-def read_zarr_string(elem):
+def read_zarr_string(elem, _reader):
     return str(elem[()])
 
 
@@ -738,7 +930,7 @@ _REGISTRY.register_read(ZarrArray, IOSpec("bytes", "0.2.0"))(read_scalar)
 
 @_REGISTRY.register_write(H5Group, np.str_, IOSpec("string", "0.2.0"))
 @_REGISTRY.register_write(H5Group, str, IOSpec("string", "0.2.0"))
-def write_string(f, k, v, dataset_kwargs):
+def write_string(f, k, v, _writer, dataset_kwargs):
     dataset_kwargs = dataset_kwargs.copy()
     dataset_kwargs.pop("compression", None)
     dataset_kwargs.pop("compression_opts", None)
