@@ -1,15 +1,23 @@
-from enum import Enum
-from functools import wraps, singledispatch
+from __future__ import annotations
+
+from functools import wraps
+from typing import TYPE_CHECKING, Callable, Literal, Union, cast
 from warnings import warn
 
-from packaging import version
 import h5py
+from packaging.version import Version
 
-from .._core.sparse_dataset import SparseDataset
+from .._core.sparse_dataset import BaseCompressedSparseDataset
+from ..compat import H5Group, ZarrGroup, add_note, pairwise
+
+if TYPE_CHECKING:
+    from .._types import StorageType
+
+    Storage = Union[StorageType, BaseCompressedSparseDataset]
 
 # For allowing h5py v3
-# https://github.com/theislab/anndata/issues/442
-H5PY_V3 = version.parse(h5py.__version__).major >= 3
+# https://github.com/scverse/anndata/issues/442
+H5PY_V3 = Version(h5py.__version__).major >= 3
 
 # -------------------------------------------------------------------------------
 # Type conversion
@@ -147,23 +155,31 @@ class AnnDataReadError(OSError):
     pass
 
 
-def _get_parent(elem):
-    try:
-        import zarr
-    except ImportError:
-        zarr = None
-    if zarr and isinstance(elem, (zarr.Group, zarr.Array)):
-        parent = elem.store  # Not sure how to always get a name out of this
-    elif isinstance(elem, SparseDataset):
-        parent = elem.group.file.name
-    else:
-        parent = elem.file.name
-    return parent
+def _get_display_path(store: Storage) -> str:
+    """Return an absolute path of an element (always starts with “/”)."""
+    if isinstance(store, BaseCompressedSparseDataset):
+        store = store.group
+    path = store.name or "??"  # can be None
+    return f'/{path.removeprefix("/")}'
+
+
+def add_key_note(
+    e: BaseException, store: Storage, path: str, key: str, op: Literal["read", "writ"]
+) -> None:
+    if any(
+        f"Error raised while {op}ing key" in note
+        for note in getattr(e, "__notes__", [])
+    ):
+        return
+
+    dir = "to" if op == "writ" else "from"
+    msg = f"Error raised while {op}ing key {key!r} of {type(store)} {dir} {path}"
+    add_note(e, msg)
 
 
 def report_read_key_on_error(func):
     """\
-    A decorator for zarr element reading which makes keys involved in errors get reported.
+    A decorator for hdf5/zarr element reading which makes keys involved in errors get reported.
 
     Example
     -------
@@ -177,25 +193,29 @@ def report_read_key_on_error(func):
     """
 
     @wraps(func)
-    def func_wrapper(elem, *args, **kwargs):
+    def func_wrapper(*args, **kwargs):
+        from anndata._io.specs import Reader
+
+        # Figure out signature (method vs function) by going through args
+        for arg in args:
+            if not isinstance(arg, Reader):
+                store = cast("Storage", arg)
+                break
+        else:
+            raise ValueError("No element found in args.")
         try:
-            return func(elem, *args, **kwargs)
+            return func(*args, **kwargs)
         except Exception as e:
-            if isinstance(e, AnnDataReadError):
-                raise e
-            else:
-                parent = _get_parent(elem)
-                raise AnnDataReadError(
-                    f"Above error raised while reading key {elem.name!r} of "
-                    f"type {type(elem)} from {parent}."
-                )
+            path, key = _get_display_path(store).rsplit("/", 1)
+            add_key_note(e, store, path or "/", key, "read")
+            raise
 
     return func_wrapper
 
 
 def report_write_key_on_error(func):
     """\
-    A decorator for zarr element reading which makes keys involved in errors get reported.
+    A decorator for hdf5/zarr element writing which makes keys involved in errors get reported.
 
     Example
     -------
@@ -209,19 +229,22 @@ def report_write_key_on_error(func):
     """
 
     @wraps(func)
-    def func_wrapper(elem, key, val, *args, **kwargs):
+    def func_wrapper(*args, **kwargs):
+        from anndata._io.specs import Writer
+
+        # Figure out signature (method vs function) by going through args
+        for arg, key in pairwise(args):
+            if not isinstance(arg, Writer):
+                store = cast("Storage", arg)
+                break
+        else:
+            raise ValueError("No element found in args.")
         try:
-            return func(elem, key, val, *args, **kwargs)
+            return func(*args, **kwargs)
         except Exception as e:
-            if "Above error raised while writing key" in format(e):
-                raise
-            else:
-                parent = _get_parent(elem)
-                raise type(e)(
-                    f"{e}\n\n"
-                    f"Above error raised while writing key {key!r} of {type(elem)} "
-                    f"to {parent}"
-                ) from e
+            path = _get_display_path(store)
+            add_key_note(e, store, path, key, "writ")
+            raise
 
     return func_wrapper
 
@@ -231,7 +254,14 @@ def report_write_key_on_error(func):
 # -------------------------------------------------------------------------------
 
 
-def _read_legacy_raw(f, modern_raw, read_df, read_attr, *, attrs=("X", "var", "varm")):
+def _read_legacy_raw(
+    f: ZarrGroup | H5Group,
+    modern_raw,  # TODO: type
+    read_df: Callable,
+    read_attr: Callable,
+    *,
+    attrs=("X", "var", "varm"),
+) -> dict:
     """\
     Backwards compat for reading legacy raw.
     Makes sure that no modern raw group coexists with legacy raw.* groups.
