@@ -1,34 +1,33 @@
-from collections.abc import MutableMapping
+from __future__ import annotations
+
 from pathlib import Path
-from typing import TypeVar, Union
+from typing import TYPE_CHECKING, TypeVar
 from warnings import warn
 
 import numpy as np
-from scipy import sparse
 import pandas as pd
 import zarr
+from scipy import sparse
 
-from .._core.anndata import AnnData
-from ..compat import (
-    _from_fixed_length_strings,
-    _clean_uns,
-)
-from .utils import (
-    report_read_key_on_error,
-    _read_legacy_raw,
-)
-from .specs import read_elem, write_elem
 from anndata._warnings import OldFormatWarning
 
+from .._core.anndata import AnnData
+from ..compat import _clean_uns, _from_fixed_length_strings
+from ..experimental import read_dispatched, write_dispatched
+from .specs import read_elem
+from .utils import _read_legacy_raw, report_read_key_on_error
+
+if TYPE_CHECKING:
+    from collections.abc import MutableMapping
 
 T = TypeVar("T")
 
 
 def write_zarr(
-    store: Union[MutableMapping, str, Path],
+    store: MutableMapping | str | Path,
     adata: AnnData,
     chunks=None,
-    **dataset_kwargs,
+    **ds_kwargs,
 ) -> None:
     if isinstance(store, Path):
         store = str(store)
@@ -39,24 +38,17 @@ def write_zarr(
     f = zarr.open(store, mode="w")
     f.attrs.setdefault("encoding-type", "anndata")
     f.attrs.setdefault("encoding-version", "0.1.0")
-    if chunks is not None and not isinstance(adata.X, sparse.spmatrix):
-        write_elem(
-            f, "X", adata.X, dataset_kwargs=dict(chunks=chunks, **dataset_kwargs)
-        )
-    else:
-        write_elem(f, "X", adata.X, dataset_kwargs=dataset_kwargs)
-    write_elem(f, "obs", adata.obs, dataset_kwargs=dataset_kwargs)
-    write_elem(f, "var", adata.var, dataset_kwargs=dataset_kwargs)
-    write_elem(f, "obsm", dict(adata.obsm), dataset_kwargs=dataset_kwargs)
-    write_elem(f, "varm", dict(adata.varm), dataset_kwargs=dataset_kwargs)
-    write_elem(f, "obsp", dict(adata.obsp), dataset_kwargs=dataset_kwargs)
-    write_elem(f, "varp", dict(adata.varp), dataset_kwargs=dataset_kwargs)
-    write_elem(f, "layers", dict(adata.layers), dataset_kwargs=dataset_kwargs)
-    write_elem(f, "uns", dict(adata.uns), dataset_kwargs=dataset_kwargs)
-    write_elem(f, "raw", adata.raw, dataset_kwargs=dataset_kwargs)
+
+    def callback(func, s, k, elem, dataset_kwargs, iospec):
+        if chunks is not None and not isinstance(elem, sparse.spmatrix) and k == "/X":
+            func(s, k, elem, dataset_kwargs=dict(chunks=chunks, **dataset_kwargs))
+        else:
+            func(s, k, elem, dataset_kwargs=dataset_kwargs)
+
+    write_dispatched(f, "/", adata, callback=callback, dataset_kwargs=ds_kwargs)
 
 
-def read_zarr(store: Union[str, Path, MutableMapping, zarr.Group]) -> AnnData:
+def read_zarr(store: str | Path | MutableMapping | zarr.Group) -> AnnData:
     """\
     Read from a hierarchical Zarr array store.
 
@@ -68,29 +60,43 @@ def read_zarr(store: Union[str, Path, MutableMapping, zarr.Group]) -> AnnData:
     if isinstance(store, Path):
         store = str(store)
 
-    f = zarr.open(store, mode="r")
+    if isinstance(store, zarr.Group):
+        f = store
+    else:
+        f = zarr.open(store, mode="r")
 
-    if "encoding-type" in f.attrs:
-        return read_elem(f[""])
+    # Read with handling for backwards compat
+    def callback(func, elem_name: str, elem, iospec):
+        if iospec.encoding_type == "anndata" or elem_name.endswith("/"):
+            return AnnData(
+                **{
+                    k: read_dispatched(v, callback)
+                    for k, v in elem.items()
+                    if not k.startswith("raw.")
+                }
+            )
+        elif elem_name.startswith("/raw."):
+            return None
+        elif elem_name in {"/obs", "/var"}:
+            return read_dataframe(elem)
+        elif elem_name == "/raw":
+            # Backwards compat
+            return _read_legacy_raw(f, func(elem), read_dataframe, func)
+        return func(elem)
 
-    # Backwards compat
-    d = {}
-    for k in f.keys():
-        if k.startswith("raw."):
-            continue
-        if k in {"obs", "var"}:
-            d[k] = read_dataframe(f[k])
-        else:  # Base case
-            d[k] = read_elem(f[k])
+    adata = read_dispatched(f, callback=callback)
 
-    d["raw"] = _read_legacy_raw(f, d.get("raw"), read_dataframe, read_elem)
+    # Backwards compat (should figure out which version)
+    if "raw.X" in f:
+        raw = AnnData(**_read_legacy_raw(f, adata.raw, read_dataframe, read_elem))
+        raw.obs_names = adata.obs_names
+        adata.raw = raw
 
-    if "X" in d:
-        d["dtype"] = d["X"].dtype
+    # Backwards compat for <0.7
+    if isinstance(f["obs"], zarr.Array):
+        _clean_uns(adata)
 
-    _clean_uns(d)
-
-    return AnnData(**d)
+    return adata
 
 
 @report_read_key_on_error
@@ -127,9 +133,7 @@ def read_dataframe_legacy(dataset: zarr.Array) -> pd.DataFrame:
 
 
 @report_read_key_on_error
-def read_dataframe(group) -> pd.DataFrame:
-    from .specs import _REGISTRY
-
+def read_dataframe(group: zarr.Group | zarr.Array) -> pd.DataFrame:
     # Fast paths
     if isinstance(group, zarr.Array):
         return read_dataframe_legacy(group)
