@@ -1,29 +1,73 @@
-from copy import deepcopy
-from functools import reduce, singledispatch, wraps
+from __future__ import annotations
+
+import os
+import sys
 from codecs import decode
-from inspect import signature, Parameter
-from typing import Any, Collection, Union, Mapping, MutableMapping, Optional
+from collections.abc import Mapping
+from contextlib import AbstractContextManager
+from dataclasses import dataclass, field
+from functools import singledispatch, wraps
+from inspect import Parameter, signature
+from pathlib import Path
+from typing import Any, Union
 from warnings import warn
 
 import h5py
-from scipy.sparse import spmatrix
 import numpy as np
 import pandas as pd
+from packaging.version import Version
+from scipy.sparse import issparse, spmatrix
 
-from ._overloaded_dict import _overloaded_uns, OverloadedDict
-from .._core.index import _subset
+from .exceptiongroups import add_note  # noqa: F401
 
 
 class Empty:
     pass
 
 
-H5Group = Union[h5py.Group, h5py.File]
+Index1D = Union[slice, int, str, np.int64, np.ndarray]
+Index = Union[Index1D, tuple[Index1D, Index1D], spmatrix]
+H5Group = h5py.Group
 H5Array = h5py.Dataset
 
 
-# try importing zarr, dask, and zappy
-from packaging import version
+#############################
+# stdlib
+#############################
+
+
+if sys.version_info >= (3, 11):
+    from contextlib import chdir
+else:
+
+    @dataclass
+    class chdir(AbstractContextManager):
+        path: Path
+        _old_cwd: list[Path] = field(default_factory=list)
+
+        def __enter__(self) -> None:
+            self._old_cwd.append(Path())
+            os.chdir(self.path)
+
+        def __exit__(self, *_exc_info) -> None:
+            os.chdir(self._old_cwd.pop())
+
+
+if sys.version_info >= (3, 10):
+    from itertools import pairwise
+else:
+
+    def pairwise(iterable):
+        from itertools import tee
+
+        a, b = tee(iterable)
+        next(b, None)
+        return zip(a, b)
+
+
+#############################
+# Optional deps
+#############################
 
 try:
     from zarr.core import Array as ZarrArray
@@ -41,6 +85,19 @@ except ImportError:
             return "mock zarr.core.Group"
 
 
+try:
+    import awkward
+
+    AwkArray = awkward.Array
+
+except ImportError:
+
+    class AwkArray:
+        @staticmethod
+        def __repr__():
+            return "mock awkward.highlevel.Array"
+
+
 try:
     from zappy.base import ZappyArray
 except ImportError:
@@ -62,24 +119,46 @@ except ImportError:
 
 
 try:
-    from typing import Literal
+    from cupy import ndarray as CupyArray
+    from cupyx.scipy.sparse import (
+        csc_matrix as CupyCSCMatrix,
+    )
+    from cupyx.scipy.sparse import (
+        csr_matrix as CupyCSRMatrix,
+    )
+    from cupyx.scipy.sparse import (
+        spmatrix as CupySparseMatrix,
+    )
 except ImportError:
-    try:
-        from typing_extensions import Literal
-    except ImportError:
 
-        class LiteralMeta(type):
-            def __getitem__(cls, values):
-                if not isinstance(values, tuple):
-                    values = (values,)
-                return type("Literal_", (Literal,), dict(__args__=values))
+    class CupySparseMatrix:
+        @staticmethod
+        def __repr__():
+            return "mock cupyx.scipy.sparse.spmatrix"
 
-        class Literal(metaclass=LiteralMeta):
-            pass
+    class CupyCSRMatrix:
+        @staticmethod
+        def __repr__():
+            return "mock cupyx.scipy.sparse.csr_matrix"
+
+    class CupyCSCMatrix:
+        @staticmethod
+        def __repr__():
+            return "mock cupyx.scipy.sparse.csc_matrix"
+
+    class CupyArray:
+        @staticmethod
+        def __repr__():
+            return "mock cupy.ndarray"
+
+
+#############################
+# IO helpers
+#############################
 
 
 @singledispatch
-def _read_attr(attrs: Mapping, name: str, default: Optional[Any] = Empty):
+def _read_attr(attrs: Mapping, name: str, default: Any | None = Empty):
     if default is Empty:
         return attrs[name]
     else:
@@ -88,7 +167,7 @@ def _read_attr(attrs: Mapping, name: str, default: Optional[Any] = Empty):
 
 @_read_attr.register(h5py.AttributeManager)
 def _read_attr_hdf5(
-    attrs: h5py.AttributeManager, name: str, default: Optional[Any] = Empty
+    attrs: h5py.AttributeManager, name: str, default: Any | None = Empty
 ):
     """
     Read an HDF5 attribute and perform all necessary conversions.
@@ -118,7 +197,7 @@ def _from_fixed_length_strings(value):
     """\
     Convert from fixed length strings to unicode.
 
-    For backwards compatability with older h5ad and zarr files.
+    For backwards compatibility with older h5ad and zarr files.
     """
     new_dtype = []
     for dt in value.dtype.descr:
@@ -139,7 +218,7 @@ def _from_fixed_length_strings(value):
 
 
 def _decode_structured_array(
-    arr: np.ndarray, dtype: Optional[np.dtype] = None, copy: bool = False
+    arr: np.ndarray, dtype: np.dtype | None = None, copy: bool = False
 ) -> np.ndarray:
     """
     h5py 3.0 now reads all strings as bytes. There is a helper method which can convert these to strings,
@@ -189,31 +268,31 @@ def _to_fixed_length_strings(value: np.ndarray) -> np.ndarray:
 #############################
 
 
-def _clean_uns(d: Mapping[str, MutableMapping[str, Union[pd.Series, str, int]]]):
+def _clean_uns(adata: AnnData):  # noqa: F821
     """
     Compat function for when categorical keys were stored in uns.
     This used to be buggy because when storing categorical columns in obs and var with
     the same column name, only one `<colname>_categories` is retained.
     """
     k_to_delete = set()
-    for cats_name, cats in d.get("uns", {}).items():
+    for cats_name, cats in adata.uns.items():
         if not cats_name.endswith("_categories"):
             continue
         name = cats_name.replace("_categories", "")
         # fix categories with a single category
         if isinstance(cats, (str, int)):
             cats = [cats]
-        for ann in ["obs", "var"]:
-            if name not in d[ann]:
+        for ann in [adata.obs, adata.var]:
+            if name not in ann:
                 continue
-            codes: np.ndarray = d[ann][name].values
+            codes: np.ndarray = ann[name].values
             # hack to maybe find the axis the categories were for
             if not np.all(codes < len(cats)):
                 continue
-            d[ann][name] = pd.Categorical.from_codes(codes, cats)
+            ann[name] = pd.Categorical.from_codes(codes, cats)
             k_to_delete.add(cats_name)
     for cats_name in k_to_delete:
-        del d["uns"][cats_name]
+        del adata.uns[cats_name]
 
 
 def _move_adj_mtx(d):
@@ -247,31 +326,6 @@ def _find_sparse_matrices(d: Mapping, n: int, keys: tuple, paths: list):
     return paths
 
 
-def _slice_uns_sparse_matrices(uns: MutableMapping, oidx: "Index1d", orig_n_obs: int):
-    """slice sparse spatrices of n_obs × n_obs in self.uns"""
-    if isinstance(oidx, slice) and len(range(*oidx.indices(orig_n_obs))) == orig_n_obs:
-        return uns  # slice of entire dimension is a no-op
-
-    paths = _find_sparse_matrices(uns, orig_n_obs, (), [])
-
-    if not paths:
-        return uns
-
-    uns = deepcopy(uns)
-    for path in paths:
-        str_path = "".join(f"['{key}']" for key in path)
-        warn(
-            f"During AnnData slicing, found matrix at .uns{str_path} that happens"
-            f" to be dimensioned at n_obs×n_obs ({orig_n_obs}×{orig_n_obs}).\n\n"
-            "These matrices should now be stored in the .obsp attribute.\n"
-            "This slicing behavior will be removed in anndata 0.8.",
-            FutureWarning,
-        )
-        d = reduce(lambda d, k: d[k], path[:-1], uns)
-        d[path[-1]] = _subset(d[path[-1]], (oidx, oidx))
-    return uns
-
-
 # This function was adapted from scikit-learn
 # github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/validation.py
 def _deprecate_positional_args(func=None, *, version: str = "1.0 (renaming of 0.25)"):
@@ -306,7 +360,7 @@ def _deprecate_positional_args(func=None, *, version: str = "1.0 (renaming of 0.
 
             # extra_args > 0
             args_msg = [
-                "{}={}".format(name, arg)
+                f"{name}={arg}"
                 for name, arg in zip(kwonly_args[:extra_args], args[-extra_args:])
             ]
             args_msg = ", ".join(args_msg)
@@ -324,3 +378,38 @@ def _deprecate_positional_args(func=None, *, version: str = "1.0 (renaming of 0.
         return _inner_deprecate_positional_args(func)
 
     return _inner_deprecate_positional_args
+
+
+def _transpose_by_block(dask_array: DaskArray) -> DaskArray:
+    import dask.array as da
+
+    b = dask_array.blocks
+    b_raveled = b.ravel()
+    block_layout = np.zeros(b.shape, dtype=object)
+
+    for i in range(block_layout.size):
+        block_layout.flat[i] = b_raveled[i].map_blocks(
+            lambda x: x.T, chunks=b_raveled[i].chunks[::-1]
+        )
+
+    return da.block(block_layout.T.tolist())
+
+
+def _safe_transpose(x):
+    """Safely transpose x
+
+    This is a workaround for: https://github.com/scipy/scipy/issues/19161
+    """
+
+    if isinstance(x, DaskArray) and issparse(x._meta):
+        return _transpose_by_block(x)
+    else:
+        return x.T
+
+
+def _map_cat_to_str(cat: pd.Categorical) -> pd.Categorical:
+    if Version(pd.__version__) >= Version("2.1"):
+        # Argument added in pandas 2.1
+        return cat.map(str, na_action="ignore")
+    else:
+        return cat.map(str)
