@@ -1,8 +1,11 @@
+from __future__ import annotations
+
 import warnings
 
 import pytest
 
 import anndata
+from anndata.tests.helpers import subset_func  # noqa: F401
 
 # TODO: Should be done in pyproject.toml, see anndata/conftest.py
 warnings.filterwarnings("ignore", category=anndata.OldFormatWarning)
