@@ -1,12 +1,13 @@
 """Test handling of values in `obs`/ `var`"""
-from natsort import natsorted
+from __future__ import annotations
+
 import numpy as np
 import pandas as pd
+import pytest
+from natsort import natsorted
 
 import anndata as ad
 
-import pytest
-
 
 @pytest.mark.parametrize("dtype", [object, "string"])
 def test_str_to_categorical(dtype):
@@ -39,7 +40,7 @@ def test_to_categorical_ordering(dtype):
 
 
 def test_non_str_to_not_categorical():
-    # Test case based on https://github.com/theislab/anndata/issues/141#issuecomment-802105259
+    # Test case based on https://github.com/scverse/anndata/issues/141#issuecomment-802105259
     obs = pd.DataFrame(index=[f"cell-{i}" for i in range(5)]).assign(
         str_with_nan=["foo", "bar", None, np.nan, "foo"],
         boolean_with_nan_and_none=[True, False, np.nan, None, True],
