@@ -1,16 +1,19 @@
+from __future__ import annotations
+
+import re
+import warnings
 from itertools import product
 
 import numpy as np
-from numpy import ma
 import pandas as pd
 import pytest
+from numpy import ma
 from scipy import sparse as sp
 from scipy.sparse import csr_matrix, issparse
 
 from anndata import AnnData
 from anndata.tests.helpers import assert_equal, gen_adata
 
-
 # some test objects that we use below
 adata_dense = AnnData(np.array([[1, 2], [3, 4]]))
 adata_dense.layers["test"] = adata_dense.X
@@ -38,9 +41,6 @@ def test_creation():
     assert adata.raw.X.tolist() == X.tolist()
     assert adata.raw.var_names.tolist() == ["a", "b", "c"]
 
-    with pytest.raises(ValueError):
-        AnnData(np.array([[1, 2], [3, 4]]), dict(TooLong=[1, 2, 3, 4]))
-
     # init with empty data matrix
     shape = (3, 5)
     adata = AnnData(None, uns=dict(test=np.array((3, 3))), shape=shape)
@@ -49,6 +49,55 @@ def test_creation():
     assert "test" in adata.uns
 
 
+@pytest.mark.parametrize(
+    ("src", "src_arg", "dim_msg"),
+    [
+        pytest.param(
+            "X",
+            adata_dense.X,
+            "`{dim}` must have as many rows as `X` has {mat_dim}s",
+            id="x",
+        ),
+        pytest.param(
+            "shape", (2, 2), "`shape` is inconsistent with `{dim}`", id="shape"
+        ),
+    ],
+)
+@pytest.mark.parametrize("dim", ["obs", "var"])
+@pytest.mark.parametrize(
+    ("dim_arg", "msg"),
+    [
+        pytest.param(
+            lambda _: dict(TooLong=[1, 2, 3, 4]),
+            "Length of values (4) does not match length of index (2)",
+            id="too_long_col",
+        ),
+        pytest.param(
+            lambda dim: {f"{dim}_names": ["a", "b", "c"]}, None, id="too_many_names"
+        ),
+        pytest.param(
+            lambda _: pd.DataFrame(index=["a", "b", "c"]), None, id="too_long_df"
+        ),
+    ],
+)
+def test_creation_error(src, src_arg, dim_msg, dim, dim_arg, msg: str | None):
+    if msg is None:
+        mat_dim = "row" if dim == "obs" else "column"
+        msg = dim_msg.format(dim=dim, mat_dim=mat_dim)
+    with pytest.raises(ValueError, match=re.escape(msg)):
+        AnnData(**{src: src_arg, dim: dim_arg(dim)})
+
+
+def test_invalid_X():
+    with pytest.raises(
+        ValueError,
+        match=re.escape(
+            "X needs to be of one of numpy.ndarray, numpy.ma.core.MaskedArray, scipy.sparse.spmatrix, h5py.Dataset, zarr.Array, anndata.experimental.[CSC,CSR]Dataset, dask.array.Array, cupy.ndarray, cupyx.scipy.sparse.spmatrix, not <class 'str'>."
+        ),
+    ):
+        AnnData("string is not a valid X")
+
+
 def test_create_with_dfs():
     X = np.ones((6, 3))
     obs = pd.DataFrame(dict(cat_anno=pd.Categorical(["a", "a", "a", "a", "b", "a"])))
@@ -94,6 +143,13 @@ def test_create_from_df_with_obs_and_var():
         AnnData(df, var=var.reset_index())
 
 
+def test_matching_int_index():
+    adata = AnnData(
+        pd.DataFrame(dict(a=[0.0, 0.5]), index=[0, 1]), obs=pd.DataFrame(index=[0, 1])
+    )
+    pd.testing.assert_index_equal(adata.obs_names, pd.Index(["0", "1"]))
+
+
 def test_from_df_and_dict():
     df = pd.DataFrame(dict(a=[0.1, 0.2, 0.3], b=[1.1, 1.2, 1.3]))
     adata = AnnData(df, dict(species=pd.Categorical(["a", "b", "a"])))
@@ -164,7 +220,7 @@ def test_setting_index_names_error(attr):
     orig = adata_sparse[:2, :2]
     adata = adata_sparse[:2, :2]
     assert getattr(adata, attr).name is None
-    with pytest.raises(ValueError, match=fr"AnnData expects \.{attr[:3]}\.index\.name"):
+    with pytest.raises(ValueError, match=rf"AnnData expects \.{attr[:3]}\.index\.name"):
         setattr(adata, attr, pd.Index(["x", "y"], name=0))
     assert adata.is_view
     assert getattr(adata, attr).tolist() != ["x", "y"]
@@ -322,21 +378,6 @@ def test_slicing_strings():
         adata[["A", "B", "not_in_obs"], :]
 
 
-def test_slicing_graphs():
-    # Testing for deprecated behaviour of connectivity matrices in .uns["neighbors"]
-    with pytest.warns(FutureWarning, match=r".obsp\['connectivities'\]"):
-        adata = AnnData(
-            np.array([[1, 2], [3, 4], [5, 6]]),
-            uns=dict(neighbors=dict(connectivities=sp.csr_matrix(np.ones((3, 3))))),
-        )
-
-    adata_sub = adata[[0, 1], :]
-    with pytest.warns(FutureWarning):
-        assert adata_sub.uns["neighbors"]["connectivities"].shape[0] == 2
-        assert adata.uns["neighbors"]["connectivities"].shape[0] == 3
-        assert adata_sub.copy().uns["neighbors"]["connectivities"].shape[0] == 2
-
-
 def test_slicing_series():
     adata = AnnData(
         np.array([[1, 2], [3, 4], [5, 6]]),
@@ -452,7 +493,9 @@ def test_rename_categories():
     adata.uns["tool"]["params"] = dict(groupby="cat_anno")
 
     new_categories = ["c", "d"]
-    adata.rename_categories("cat_anno", new_categories)
+    with warnings.catch_warnings():
+        warnings.simplefilter("error")
+        adata.rename_categories("cat_anno", new_categories)
 
     assert list(adata.obs["cat_anno"].cat.categories) == new_categories
     assert list(adata.uns["tool"]["cat_array"].dtype.names) == new_categories
@@ -604,3 +647,21 @@ def test_copy():
         assert_eq_not_id(map_sprs.keys(), map_copy.keys())
         for key in map_sprs.keys():
             assert_eq_not_id(map_sprs[key], map_copy[key])
+
+
+def test_to_memory_no_copy():
+    adata = gen_adata((3, 5))
+    mem = adata.to_memory()
+
+    assert mem.X is adata.X
+    # Currently does not hold for `obs`, `var`, but should in future
+    for key in adata.layers:
+        assert mem.layers[key] is adata.layers[key]
+    for key in adata.obsm:
+        assert mem.obsm[key] is adata.obsm[key]
+    for key in adata.varm:
+        assert mem.varm[key] is adata.varm[key]
+    for key in adata.obsp:
+        assert mem.obsp[key] is adata.obsp[key]
+    for key in adata.varp:
+        assert mem.varp[key] is adata.varp[key]
