@@ -1,25 +1,42 @@
-from collections.abc import Hashable
-from copy import deepcopy
-from itertools import chain, product
-from functools import partial, singledispatch
+from __future__ import annotations
+
 import warnings
+from collections.abc import Hashable
+from contextlib import nullcontext
+from copy import deepcopy
+from functools import partial, singledispatch
+from itertools import chain, permutations, product
+from typing import Any, Callable, Literal
 
 import numpy as np
-from numpy import ma
 import pandas as pd
-from pandas.api.types import is_categorical_dtype
 import pytest
+from boltons.iterutils import default_exit, remap, research
+from numpy import ma
+from packaging.version import Version
 from scipy import sparse
-from boltons.iterutils import research, remap, default_exit
-
 
 from anndata import AnnData, Raw, concat
-from anndata._core.index import _subset
 from anndata._core import merge
+from anndata._core.index import _subset
+from anndata.compat import AwkArray, DaskArray
 from anndata.tests import helpers
-from anndata.tests.helpers import assert_equal, gen_adata
+from anndata.tests.helpers import (
+    BASE_MATRIX_PARAMS,
+    CUPY_MATRIX_PARAMS,
+    DASK_MATRIX_PARAMS,
+    GEN_ADATA_DASK_ARGS,
+    as_dense_dask_array,
+    assert_equal,
+    gen_adata,
+    gen_vstr_recarray,
+)
 from anndata.utils import asarray
 
+mark_legacy_concatenate = pytest.mark.filterwarnings(
+    r"ignore:.*AnnData\.concatenate is deprecated:FutureWarning"
+)
+
 
 @singledispatch
 def filled_like(a, fill_value=None):
@@ -27,12 +44,17 @@ def filled_like(a, fill_value=None):
 
 
 @filled_like.register(np.ndarray)
-def _filled_array(a, fill_value=None):
+def _filled_array_np(a, fill_value=None):
     if fill_value is None:
         fill_value = np.nan
     return np.broadcast_to(fill_value, a.shape)
 
 
+@filled_like.register(DaskArray)
+def _filled_array(a, fill_value=None):
+    return as_dense_dask_array(_filled_array_np(a, fill_value))
+
+
 @filled_like.register(sparse.spmatrix)
 def _filled_sparse(a, fill_value=None):
     if fill_value is None:
@@ -60,10 +82,9 @@ def make_idx_tuple(idx, axis):
     return tuple(tup)
 
 
-@pytest.fixture(
-    params=[asarray, sparse.csr_matrix, sparse.csc_matrix],
-    ids=["np_array", "scipy_csr", "scipy_csc"],
-)
+# Will call func(sparse_matrix) so these types should be sparse compatible
+# See array_type if only dense arrays are expected as input.
+@pytest.fixture(params=BASE_MATRIX_PARAMS + DASK_MATRIX_PARAMS + CUPY_MATRIX_PARAMS)
 def array_type(request):
     return request.param
 
@@ -79,7 +100,7 @@ def fill_val(request):
 
 
 @pytest.fixture(params=[0, 1])
-def axis(request):
+def axis(request) -> Literal[0, 1]:
     return request.param
 
 
@@ -114,7 +135,7 @@ def fix_known_differences(orig, result, backwards_compat=True):
 
     # Possibly need to fix this, ordered categoricals lose orderedness
     for k, dtype in orig.obs.dtypes.items():
-        if is_categorical_dtype(dtype) and dtype.ordered:
+        if isinstance(dtype, pd.CategoricalDtype) and dtype.ordered:
             result.obs[k] = result.obs[k].astype(dtype)
 
     return orig, result
@@ -131,6 +152,7 @@ def test_concat_interface_errors():
         concat([])
 
 
+@mark_legacy_concatenate
 @pytest.mark.parametrize(
     ["concat_func", "backwards_compat"],
     [
@@ -139,7 +161,7 @@ def test_concat_interface_errors():
     ],
 )
 def test_concatenate_roundtrip(join_type, array_type, concat_func, backwards_compat):
-    adata = gen_adata((100, 10), X_type=array_type)
+    adata = gen_adata((100, 10), X_type=array_type, **GEN_ADATA_DASK_ARGS)
 
     remaining = adata.obs_names
     subsets = []
@@ -159,6 +181,7 @@ def test_concatenate_roundtrip(join_type, array_type, concat_func, backwards_com
     assert_equal(result[orig.obs_names].copy(), orig)
 
 
+@mark_legacy_concatenate
 def test_concatenate_dense():
     # dense data
     X1 = np.array([[1, 2, 3], [4, 5, 6]])
@@ -234,6 +257,7 @@ def test_concatenate_dense():
     assert np.allclose(var_ma.compressed(), var_ma_ref.compressed())
 
 
+@mark_legacy_concatenate
 def test_concatenate_layers(array_type, join_type):
     adatas = []
     for _ in range(5):
@@ -293,6 +317,7 @@ def obsm_adatas():
     ]
 
 
+@mark_legacy_concatenate
 def test_concatenate_obsm_inner(obsm_adatas):
     adata = obsm_adatas[0].concatenate(obsm_adatas[1:], join="inner")
 
@@ -322,6 +347,7 @@ def test_concatenate_obsm_inner(obsm_adatas):
     pd.testing.assert_frame_equal(true_df, cur_df)
 
 
+@mark_legacy_concatenate
 def test_concatenate_obsm_outer(obsm_adatas, fill_val):
     outer = obsm_adatas[0].concatenate(
         obsm_adatas[1:], join="outer", fill_value=fill_val
@@ -392,6 +418,7 @@ def test_concat_annot_join(obsm_adatas, join_type):
     )
 
 
+@mark_legacy_concatenate
 def test_concatenate_layers_misaligned(array_type, join_type):
     adatas = []
     for _ in range(5):
@@ -405,6 +432,7 @@ def test_concatenate_layers_misaligned(array_type, join_type):
     assert_equal(merged.X, merged.layers["a"])
 
 
+@mark_legacy_concatenate
 def test_concatenate_layers_outer(array_type, fill_val):
     # Testing that issue #368 is fixed
     a = AnnData(
@@ -420,6 +448,7 @@ def test_concatenate_layers_outer(array_type, fill_val):
     )
 
 
+@mark_legacy_concatenate
 def test_concatenate_fill_value(fill_val):
     def get_obs_els(adata):
         return {
@@ -430,20 +459,27 @@ def test_concatenate_fill_value(fill_val):
 
     adata1 = gen_adata((10, 10))
     adata1.obsm = {
-        k: v for k, v in adata1.obsm.items() if not isinstance(v, pd.DataFrame)
+        k: v
+        for k, v in adata1.obsm.items()
+        if not isinstance(v, (pd.DataFrame, AwkArray))
     }
     adata2 = gen_adata((10, 5))
     adata2.obsm = {
         k: v[:, : v.shape[1] // 2]
         for k, v in adata2.obsm.items()
-        if not isinstance(v, pd.DataFrame)
+        if not isinstance(v, (pd.DataFrame, AwkArray))
     }
     adata3 = gen_adata((7, 3))
     adata3.obsm = {
         k: v[:, : v.shape[1] // 3]
         for k, v in adata3.obsm.items()
-        if not isinstance(v, pd.DataFrame)
+        if not isinstance(v, (pd.DataFrame, AwkArray))
     }
+    # remove AwkArrays from adata.var, as outer joins are not yet implemented for them
+    for tmp_ad in [adata1, adata2, adata3]:
+        for k in [k for k, v in tmp_ad.varm.items() if isinstance(v, AwkArray)]:
+            del tmp_ad.varm[k]
+
     joined = adata1.concatenate([adata2, adata3], join="outer", fill_value=fill_val)
 
     ptr = 0
@@ -458,6 +494,7 @@ def test_concatenate_fill_value(fill_val):
         ptr += orig.n_obs
 
 
+@mark_legacy_concatenate
 def test_concatenate_dense_duplicates():
     X1 = np.array([[1, 2, 3], [4, 5, 6]])
     X2 = np.array([[1, 2, 3], [4, 5, 6]])
@@ -509,6 +546,7 @@ def test_concatenate_dense_duplicates():
     ]
 
 
+@mark_legacy_concatenate
 def test_concatenate_sparse():
     # sparse data
     from scipy.sparse import csr_matrix
@@ -554,6 +592,7 @@ def test_concatenate_sparse():
     ]
 
 
+@mark_legacy_concatenate
 def test_concatenate_mixed():
     X1 = sparse.csr_matrix(np.array([[1, 2, 0], [4, 0, 6], [0, 0, 9]]))
     X2 = sparse.csr_matrix(np.array([[0, 2, 3], [4, 0, 0], [7, 0, 9]]))
@@ -589,6 +628,7 @@ def test_concatenate_mixed():
     assert isinstance(adata_all.layers["counts"], sparse.csr_matrix)
 
 
+@mark_legacy_concatenate
 def test_concatenate_with_raw():
     # dense data
     X1 = np.array([[1, 2, 3], [4, 5, 6]])
@@ -666,6 +706,100 @@ def test_concatenate_with_raw():
     assert adata_all.raw is None
 
 
+def test_concatenate_awkward(join_type):
+    import awkward as ak
+
+    a = ak.Array([[{"a": 1, "b": "foo"}], [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}]])
+    b = ak.Array(
+        [
+            [{"a": 4}, {"a": 5}],
+            [{"a": 6}],
+            [{"a": 7}],
+        ]
+    )
+
+    adata_a = AnnData(np.zeros((2, 0), dtype=float), obsm={"awk": a})
+    adata_b = AnnData(np.zeros((3, 0), dtype=float), obsm={"awk": b})
+
+    if join_type == "inner":
+        expected = ak.Array(
+            [
+                [{"a": 1}],
+                [{"a": 2}, {"a": 3}],
+                [{"a": 4}, {"a": 5}],
+                [{"a": 6}],
+                [{"a": 7}],
+            ]
+        )
+    elif join_type == "outer":
+        # TODO: This is what we would like to return, but waiting on:
+        # * https://github.com/scikit-hep/awkward/issues/2182 and awkward 2.1.0
+        # * https://github.com/scikit-hep/awkward/issues/2173
+        # expected = ak.Array(
+        #     [
+        #         [{"a": 1, "b": "foo"}],
+        #         [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}],
+        #         [{"a": 4, "b": None}, {"a": 5, "b": None}],
+        #         [{"a": 6, "b": None}],
+        #         [{"a": 7, "b": None}],
+        #     ]
+        # )
+        expected = ak.concatenate(
+            [  # I don't think I can construct a UnionArray directly
+                ak.Array(
+                    [
+                        [{"a": 1, "b": "foo"}],
+                        [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}],
+                    ]
+                ),
+                ak.Array(
+                    [
+                        [{"a": 4}, {"a": 5}],
+                        [{"a": 6}],
+                        [{"a": 7}],
+                    ]
+                ),
+            ]
+        )
+
+    result = concat([adata_a, adata_b], join=join_type).obsm["awk"]
+
+    assert_equal(expected, result)
+
+
+@pytest.mark.parametrize(
+    "other",
+    [
+        pd.DataFrame({"a": [4, 5, 6], "b": ["foo", "bar", "baz"]}, index=list("cde")),
+        np.ones((3, 2)),
+        sparse.random(3, 100, format="csr"),
+    ],
+)
+def test_awkward_does_not_mix(join_type, other):
+    import awkward as ak
+
+    awk = ak.Array(
+        [[{"a": 1, "b": "foo"}], [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}]]
+    )
+
+    adata_a = AnnData(
+        np.zeros((2, 3), dtype=float),
+        obs=pd.DataFrame(index=list("ab")),
+        obsm={"val": awk},
+    )
+    adata_b = AnnData(
+        np.zeros((3, 3), dtype=float),
+        obs=pd.DataFrame(index=list("cde")),
+        obsm={"val": other},
+    )
+
+    with pytest.raises(
+        NotImplementedError,
+        match="Cannot concatenate an AwkwardArray with other array types",
+    ):
+        concat([adata_a, adata_b], join=join_type)
+
+
 def test_pairwise_concat(axis, array_type):
     dim_sizes = [[100, 200, 50], [50, 50, 50]]
     if axis:
@@ -699,11 +833,15 @@ def test_pairwise_concat(axis, array_type):
 
     # Check values of included elements
     full_inds = np.arange(w_pairwise.shape[axis])
-    groups = getattr(w_pairwise, dim).groupby("orig").indices
+    obs_var: pd.DataFrame = getattr(w_pairwise, dim)
+    groups = obs_var.groupby("orig", observed=True).indices
     for k, inds in groups.items():
         orig_arr = getattr(adatas[k], dim_attr)["arr"]
         full_arr = getattr(w_pairwise, dim_attr)["arr"]
 
+        if isinstance(full_arr, DaskArray):
+            full_arr = full_arr.compute()
+
         # Check original values are intact
         assert_equal(orig_arr, _subset(full_arr, (inds, inds)))
         # Check that entries are filled with zeroes
@@ -846,12 +984,12 @@ def map_values(mapping, path, key, old_parent, new_parent, new_items):
     return ret
 
 
-def permute_nested_values(dicts: "List[dict]", gen_val: "Callable[[int], Any]"):
+def permute_nested_values(dicts: list[dict], gen_val: Callable[[int], Any]):
     """
     This function permutes the values of a nested mapping, for testing that out merge
     method work regardless of the values types.
 
-    Assumes the intial dictionary had integers for values.
+    Assumes the initial dictionary had integers for values.
     """
     dicts = deepcopy(dicts)
     initial_values = [
@@ -882,6 +1020,15 @@ def gen_something(n):
     return np.random.choice(options)(n)
 
 
+def gen_3d_numeric_array(n):
+    return np.random.randn(n, n, n)
+
+
+def gen_3d_recarray(_):
+    # Ignoring n as it can get quite slow
+    return gen_vstr_recarray(8, 3).reshape(2, 2, 2)
+
+
 def gen_concat_params(unss, compat2result):
     value_generators = [
         lambda x: x,
@@ -890,6 +1037,8 @@ def gen_concat_params(unss, compat2result):
         gen_list,
         gen_sparse,
         gen_something,
+        gen_3d_numeric_array,
+        gen_3d_recarray,
     ]
     for gen, (mode, result) in product(value_generators, compat2result.items()):
         yield pytest.param(unss, mode, result, gen)
@@ -971,16 +1120,14 @@ def test_concatenate_uns(unss, merge_strategy, result, value_gen):
     print(merge_strategy, "\n", unss, "\n", result)
     result, *unss = permute_nested_values([result] + unss, value_gen)
     adatas = [uns_ad(uns) for uns in unss]
-    assert_equal(
-        adatas[0].concatenate(adatas[1:], uns_merge=merge_strategy).uns,
-        result,
-        elem_name="uns",
-    )
+    with pytest.warns(FutureWarning, match=r"concatenate is deprecated"):
+        merged = AnnData.concatenate(*adatas, uns_merge=merge_strategy).uns
+    assert_equal(merged, result, elem_name="uns")
 
 
 def test_transposed_concat(array_type, axis, join_type, merge_strategy, fill_val):
-    lhs = gen_adata((10, 10), X_type=array_type)
-    rhs = gen_adata((10, 12), X_type=array_type)
+    lhs = gen_adata((10, 10), X_type=array_type, **GEN_ADATA_DASK_ARGS)
+    rhs = gen_adata((10, 12), X_type=array_type, **GEN_ADATA_DASK_ARGS)
 
     a = concat([lhs, rhs], axis=axis, join=join_type, merge=merge_strategy)
     b = concat(
@@ -990,14 +1137,14 @@ def test_transposed_concat(array_type, axis, join_type, merge_strategy, fill_val
     assert_equal(a, b)
 
 
-def test_batch_key(axis):
+def test_batch_key(axis, array_type):
     """Test that concat only adds a label if the key is provided"""
 
     def get_annot(adata):
         return getattr(adata, ("obs", "var")[axis])
 
-    lhs = gen_adata((10, 10))
-    rhs = gen_adata((10, 12))
+    lhs = gen_adata((10, 10), **GEN_ADATA_DASK_ARGS)
+    rhs = gen_adata((10, 12), **GEN_ADATA_DASK_ARGS)
 
     # There is probably a prettier way to do this
     annot = get_annot(concat([lhs, rhs], axis=axis))
@@ -1038,6 +1185,125 @@ def test_concat_categories_from_mapping():
     )
 
 
+def test_concat_categories_maintain_dtype():
+    a = AnnData(
+        X=np.ones((5, 1)),
+        obs=pd.DataFrame(
+            {
+                "cat": pd.Categorical(list("aabcc")),
+                "cat_ordered": pd.Categorical(list("aabcc"), ordered=True),
+            },
+            index=[f"cell{i:02}" for i in range(5)],
+        ),
+    )
+    b = AnnData(
+        X=np.ones((5, 1)),
+        obs=pd.DataFrame(
+            {
+                "cat": pd.Categorical(list("bccdd")),
+                "cat_ordered": pd.Categorical(list("bccdd"), ordered=True),
+            },
+            index=[f"cell{i:02}" for i in range(5, 10)],
+        ),
+    )
+    c = AnnData(
+        X=np.ones((5, 1)),
+        obs=pd.DataFrame(
+            {
+                "cat_ordered": pd.Categorical(list("bccdd"), ordered=True),
+            },
+            index=[f"cell{i:02}" for i in range(5, 10)],
+        ),
+    )
+
+    result = concat({"a": a, "b": b, "c": c}, join="outer")
+
+    assert isinstance(
+        result.obs["cat"].dtype, pd.CategoricalDtype
+    ), f"Was {result.obs['cat'].dtype}"
+    assert pd.api.types.is_string_dtype(result.obs["cat_ordered"])
+
+
+def test_concat_ordered_categoricals_retained():
+    a = AnnData(
+        X=np.ones((5, 1)),
+        obs=pd.DataFrame(
+            {
+                "cat_ordered": pd.Categorical(list("aabcd"), ordered=True),
+            },
+            index=[f"cell{i:02}" for i in range(5)],
+        ),
+    )
+    b = AnnData(
+        X=np.ones((5, 1)),
+        obs=pd.DataFrame(
+            {
+                "cat_ordered": pd.Categorical(list("abcdd"), ordered=True),
+            },
+            index=[f"cell{i:02}" for i in range(5, 10)],
+        ),
+    )
+
+    c = concat([a, b])
+
+    assert isinstance(c.obs["cat_ordered"].dtype, pd.CategoricalDtype)
+    assert c.obs["cat_ordered"].cat.ordered
+
+
+def test_concat_categorical_dtype_promotion():
+    """https://github.com/scverse/anndata/issues/1170
+
+    When concatenating categorical with other dtype, defer to pandas.
+    """
+    a = AnnData(
+        np.ones((3, 3)),
+        obs=pd.DataFrame(
+            {"col": pd.Categorical(["a", "a", "b"])},
+            index=[f"cell_{i:02d}" for i in range(3)],
+        ),
+    )
+    b = AnnData(
+        np.ones((3, 3)),
+        obs=pd.DataFrame(
+            {"col": ["c", "c", "c"]},
+            index=[f"cell_{i:02d}" for i in range(3, 6)],
+        ),
+    )
+
+    result = concat([a, b])
+    expected = pd.concat([a.obs, b.obs])
+
+    assert_equal(result.obs, expected)
+
+
+def test_bool_promotion():
+    np_bool = AnnData(
+        np.ones((5, 1)),
+        obs=pd.DataFrame({"bool": [True] * 5}, index=[f"cell{i:02}" for i in range(5)]),
+    )
+    missing = AnnData(
+        np.ones((5, 1)),
+        obs=pd.DataFrame(index=[f"cell{i:02}" for i in range(5, 10)]),
+    )
+    result = concat({"np_bool": np_bool, "b": missing}, join="outer", label="batch")
+
+    assert pd.api.types.is_bool_dtype(result.obs["bool"])
+    assert pd.isnull(result.obs.loc[result.obs["batch"] == "missing", "bool"]).all()
+
+    # Check that promotion doesn't occur if it doesn't need to:
+    np_bool_2 = AnnData(
+        np.ones((5, 1)),
+        obs=pd.DataFrame(
+            {"bool": [True] * 5}, index=[f"cell{i:02}" for i in range(5, 10)]
+        ),
+    )
+    result = concat(
+        {"np_bool": np_bool, "np_bool_2": np_bool_2}, join="outer", label="batch"
+    )
+
+    assert result.obs["bool"].dtype == np.dtype(bool)
+
+
 def test_concat_names(axis):
     def get_annot(adata):
         return getattr(adata, ("obs", "var")[axis])
@@ -1072,21 +1338,31 @@ def expected_shape(a, b, axis, join):
     "shape", [pytest.param((8, 0), id="no_var"), pytest.param((0, 10), id="no_obs")]
 )
 def test_concat_size_0_dim(axis, join_type, merge_strategy, shape):
-    # https://github.com/theislab/anndata/issues/526
+    # https://github.com/scverse/anndata/issues/526
     a = gen_adata((5, 7))
     b = gen_adata(shape)
     alt_axis = 1 - axis
     dim = ("obs", "var")[axis]
 
     expected_size = expected_shape(a, b, axis=axis, join=join_type)
-    result = concat(
-        {"a": a, "b": b},
-        axis=axis,
-        join=join_type,
-        merge=merge_strategy,
-        pairwise=True,
-        index_unique="-",
+
+    ctx_concat_empty = (
+        pytest.warns(
+            FutureWarning,
+            match=r"The behavior of DataFrame concatenation with empty or all-NA entries is deprecated",
+        )
+        if shape[axis] == 0 and Version(pd.__version__) >= Version("2.1")
+        else nullcontext()
     )
+    with ctx_concat_empty:
+        result = concat(
+            {"a": a, "b": b},
+            axis=axis,
+            join=join_type,
+            merge=merge_strategy,
+            pairwise=True,
+            index_unique="-",
+        )
     assert result.shape == expected_size
 
     if join_type == "outer":
@@ -1123,10 +1399,10 @@ def test_concat_size_0_dim(axis, join_type, merge_strategy, shape):
                     )
 
 
-@pytest.mark.parametrize("elem", ["sparse", "array", "df"])
+@pytest.mark.parametrize("elem", ["sparse", "array", "df", "da"])
 def test_concat_outer_aligned_mapping(elem):
-    a = gen_adata((5, 5))
-    b = gen_adata((3, 5))
+    a = gen_adata((5, 5), **GEN_ADATA_DASK_ARGS)
+    b = gen_adata((3, 5), **GEN_ADATA_DASK_ARGS)
     del b.obsm[elem]
 
     concated = concat({"a": a, "b": b}, join="outer", label="group")
@@ -1135,8 +1411,9 @@ def test_concat_outer_aligned_mapping(elem):
     check_filled_like(result, elem_name=f"obsm/{elem}")
 
 
+@mark_legacy_concatenate
 def test_concatenate_size_0_dim():
-    # https://github.com/theislab/anndata/issues/526
+    # https://github.com/scverse/anndata/issues/526
 
     a = gen_adata((5, 10))
     b = gen_adata((5, 0))
@@ -1161,13 +1438,11 @@ def test_concat_null_X():
     assert_equal(no_X, orig)
 
 
-# https://github.com/theislab/ehrapy/issues/151#issuecomment-1016753744
+# https://github.com/scverse/ehrapy/issues/151#issuecomment-1016753744
 def test_concat_X_dtype():
-    adatas_orig = {
-        k: AnnData(np.ones((20, 10), dtype=np.int8), dtype=np.int8) for k in list("abc")
-    }
+    adatas_orig = {k: AnnData(np.ones((20, 10), dtype=np.int8)) for k in list("abc")}
     for adata in adatas_orig.values():
-        adata.raw = AnnData(np.ones((20, 30), dtype=np.float64), dtype=np.float64)
+        adata.raw = AnnData(np.ones((20, 30), dtype=np.float64))
 
     result = concat(adatas_orig, index_unique="-")
 
@@ -1179,3 +1454,131 @@ def test_concat_X_dtype():
 # def test_concatenate_uns_types():
 #     from anndata._core.merge import UNS_STRATEGIES, UNS_STRATEGIES_TYPE
 #     assert set(UNS_STRATEGIES.keys()) == set(UNS_STRATEGIES_TYPE.__args__)
+
+
+# Tests how dask plays with other types on concatenation.
+def test_concat_different_types_dask(merge_strategy, array_type):
+    import dask.array as da
+    from scipy import sparse
+
+    import anndata as ad
+
+    varm_array = sparse.random(5, 20, density=0.5, format="csr")
+
+    ad1 = ad.AnnData(X=np.ones((5, 5)), varm={"a": varm_array})
+    ad1_other = ad.AnnData(X=np.ones((5, 5)), varm={"a": array_type(varm_array)})
+    ad2 = ad.AnnData(X=np.zeros((5, 5)), varm={"a": da.ones(5, 20)})
+
+    result1 = ad.concat([ad1, ad2], merge=merge_strategy)
+    target1 = ad.concat([ad1_other, ad2], merge=merge_strategy)
+    result2 = ad.concat([ad2, ad1], merge=merge_strategy)
+    target2 = ad.concat([ad2, ad1_other], merge=merge_strategy)
+
+    assert_equal(result1, target1)
+    assert_equal(result2, target2)
+
+
+def test_outer_concat_with_missing_value_for_df():
+    # https://github.com/scverse/anndata/issues/901
+    # TODO: Extend this test to cover all cases of missing values
+    # TODO: Check values
+    a_idx = ["a", "b", "c", "d", "e"]
+    b_idx = ["f", "g", "h", "i", "j", "k", "l", "m"]
+    a = AnnData(
+        np.ones((5, 5)),
+        obs=pd.DataFrame(index=a_idx),
+    )
+    b = AnnData(
+        np.zeros((8, 9)),
+        obs=pd.DataFrame(index=b_idx),
+        obsm={"df": pd.DataFrame({"col": np.arange(8)}, index=b_idx)},
+    )
+
+    concat([a, b], join="outer")
+
+
+def test_outer_concat_outputs_nullable_bool_writable(tmp_path):
+    a = gen_adata((5, 5), obsm_types=(pd.DataFrame,))
+    b = gen_adata((3, 5), obsm_types=(pd.DataFrame,))
+
+    del b.obsm["df"]
+
+    adatas = concat({"a": a, "b": b}, join="outer", label="group")
+    adatas.write(tmp_path / "test.h5ad")
+
+
+def test_concat_duplicated_columns(join_type):
+    # https://github.com/scverse/anndata/issues/483
+    a = AnnData(
+        obs=pd.DataFrame(
+            np.ones((5, 2)), columns=["a", "a"], index=[str(x) for x in range(5)]
+        )
+    )
+    b = AnnData(
+        obs=pd.DataFrame(
+            np.ones((5, 1)), columns=["a"], index=[str(x) for x in range(5, 10)]
+        )
+    )
+
+    with pytest.raises(pd.errors.InvalidIndexError, match=r"'a'"):
+        concat([a, b], join=join_type)
+
+
+@pytest.mark.gpu
+def test_error_on_mixed_device():
+    """https://github.com/scverse/anndata/issues/1083"""
+    import cupy
+    import cupyx.scipy.sparse as cupy_sparse
+
+    cp_adata = AnnData(
+        cupy.random.randn(10, 10),
+        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10)]),
+    )
+    cp_sparse_adata = AnnData(
+        cupy_sparse.random(10, 10, format="csr", density=0.2),
+        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10, 20)]),
+    )
+    np_adata = AnnData(
+        np.random.randn(10, 10),
+        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(20, 30)]),
+    )
+    sparse_adata = AnnData(
+        sparse.random(10, 10, format="csr", density=0.2),
+        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(30, 40)]),
+    )
+
+    adatas = {
+        "cupy": cp_adata,
+        "cupy_sparse": cp_sparse_adata,
+        "numpy": np_adata,
+        "sparse": sparse_adata,
+    }
+
+    for p in map(dict, permutations(adatas.items())):
+        print(list(p.keys()))
+        with pytest.raises(
+            NotImplementedError, match="Cannot concatenate a cupy array with other"
+        ):
+            concat(p)
+
+    for p in permutations([cp_adata, cp_sparse_adata]):
+        concat(p)
+
+
+def test_concat_on_var_outer_join(array_type):
+    # https://github.com/scverse/anndata/issues/1286
+    a = AnnData(
+        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10)]),
+        var=pd.DataFrame(index=[f"gene_{i:02d}" for i in range(10)]),
+        layers={
+            "X": array_type(np.ones((10, 10))),
+        },
+    )
+    b = AnnData(
+        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10)]),
+        var=pd.DataFrame(index=[f"gene_{i:02d}" for i in range(10, 20)]),
+    )
+
+    # This shouldn't error
+    # TODO: specify expected result while accounting for null value
+    _ = concat([a, b], join="outer", axis=1)
