@@ -1,12 +1,23 @@
+from __future__ import annotations
+
 from string import ascii_letters
 
+import numpy as np
 import pandas as pd
 import pytest
-import numpy as np
 from scipy import sparse
 
 import anndata as ad
-from anndata.tests.helpers import assert_equal, report_name, gen_adata
+from anndata.compat import add_note
+from anndata.tests.helpers import (
+    asarray,
+    assert_equal,
+    gen_adata,
+    gen_awkward,
+    pytest_8_raises,
+    report_name,
+)
+from anndata.utils import dim_len
 
 # Testing to see if all error types can have the key name appended.
 # Currently fails for 22/118 since they have required arguments. Not sure what to do about that.
@@ -40,10 +51,37 @@ def reusable_adata():
     return gen_adata((10, 10))
 
 
+@pytest.mark.parametrize(
+    "shape, datashape",
+    [
+        [(4, 2), "4 * 2 * int32"],
+        [(100, 200, None), "100 * 200 * var * int32"],
+        [(4, None), "4 * var * int32"],
+        [(0, 4), "0 * 4 * int32"],
+        [(4, 0), "4 * 0 * int32"],
+        [(8, None, None), "8 * var * var * int32"],
+        [(8, None, None, None), "8 * var * var * var * int32"],
+        [(4, None, 8), "4 * var * 8 * int32"],
+        [(100, 200, 4), "100 * 200 * 4 * int32"],
+        [(4, 0, 0), "4 * 0 * 0 * int32"],
+        [(0, 0, 0), "0 * 0 * 0 * int32"],
+        [(0, None), "0 * var * int32"],
+    ],
+)
+def test_gen_awkward(shape, datashape):
+    import awkward as ak
+
+    arr = gen_awkward(shape)
+    for i, s in enumerate(shape):
+        assert dim_len(arr, i) == s
+    arr_type = ak.types.from_datashape(datashape)
+    assert arr.type == arr_type
+
+
 # Does this work for every warning?
 def test_report_name():
     def raise_error():
-        raise Exception("an error occured!")
+        raise Exception("an error occurred!")
 
     letters = np.array(list(ascii_letters))
     tag = "".join(np.random.permutation(letters))
@@ -133,7 +171,7 @@ def test_assert_equal_raw():
 
 def test_assert_equal_raw_presence():
     # This was causing some testing issues during
-    # https://github.com/theislab/anndata/pull/542
+    # https://github.com/scverse/anndata/pull/542
     a = gen_adata((10, 20))
     b = a.copy()
     a.raw = a.copy()
@@ -188,3 +226,54 @@ def test_assert_equal_aligned_mapping_empty():
         with pytest.raises(AssertionError):
             assert_equal(getattr(adata, attr), getattr(diff_idx, attr))
         assert_equal(getattr(adata, attr), getattr(same_idx, attr))
+
+
+def test_assert_equal_dask_arrays():
+    import dask.array as da
+
+    a = da.from_array([[1, 2, 3], [4, 5, 6]])
+    b = da.from_array([[1, 2, 3], [4, 5, 6]])
+
+    assert_equal(a, b)
+
+    c = da.ones(10, dtype="int32")
+    d = da.ones(10, dtype="int64")
+    assert_equal(c, d)
+
+
+def test_assert_equal_dask_sparse_arrays():
+    import dask.array as da
+    from scipy import sparse
+
+    x = sparse.random(10, 10, format="csr", density=0.1)
+    y = da.from_array(asarray(x))
+
+    assert_equal(x, y)
+    assert_equal(y, x)
+
+
+@pytest.mark.parametrize(
+    "error, match",
+    [
+        (Exception("test"), "test"),
+        (add_note(AssertionError("foo"), "bar"), "bar"),
+        (add_note(add_note(AssertionError("foo"), "bar"), "baz"), "bar"),
+        (add_note(add_note(AssertionError("foo"), "bar"), "baz"), "baz"),
+    ],
+)
+def test_check_error_notes_success(error, match):
+    with pytest_8_raises(Exception, match=match):
+        raise error
+
+
+@pytest.mark.parametrize(
+    "error, match",
+    [
+        (Exception("test"), "foo"),
+        (add_note(AssertionError("foo"), "bar"), "baz"),
+    ],
+)
+def test_check_error_notes_failure(error, match):
+    with pytest.raises(AssertionError):
+        with pytest_8_raises(Exception, match=match):
+            raise error
