@@ -1,23 +1,33 @@
+from __future__ import annotations
+
+import re
+import warnings
+from contextlib import contextmanager
 from importlib.util import find_spec
-from os import PathLike
 from pathlib import Path
 from string import ascii_letters
-import tempfile
-import warnings
+from typing import TYPE_CHECKING
 
 import h5py
 import numpy as np
 import pandas as pd
-from pandas.api.types import is_categorical_dtype
 import pytest
-from scipy.sparse import csr_matrix, csc_matrix
 import zarr
+from numba.core.errors import NumbaDeprecationWarning
+from scipy.sparse import csc_matrix, csr_matrix
 
 import anndata as ad
-from anndata.utils import asarray
-from anndata.compat import _read_attr
+from anndata._io.specs.registry import IORegistryError
+from anndata.compat import DaskArray, _read_attr
+from anndata.tests.helpers import (
+    as_dense_dask_array,
+    assert_equal,
+    gen_adata,
+    pytest_8_raises,
+)
 
-from anndata.tests.helpers import gen_adata, assert_equal
+if TYPE_CHECKING:
+    from os import PathLike
 
 HERE = Path(__file__).parent
 
@@ -79,7 +89,7 @@ def rw(backing_h5ad):
     M, N = 100, 101
     orig = gen_adata((M, N))
     orig.write(backing_h5ad)
-    curr = ad.read(backing_h5ad)
+    curr = ad.read_h5ad(backing_h5ad)
     return curr, orig
 
 
@@ -96,7 +106,7 @@ diskfmt2 = diskfmt
 # ------------------------------------------------------------------------------
 
 
-@pytest.mark.parametrize("typ", [np.array, csr_matrix])
+@pytest.mark.parametrize("typ", [np.array, csr_matrix, as_dense_dask_array])
 def test_readwrite_roundtrip(typ, tmp_path, diskfmt, diskfmt2):
     tmpdir = Path(tmp_path)
     pth1 = tmpdir / f"first.{diskfmt}"
@@ -117,78 +127,60 @@ def test_readwrite_roundtrip(typ, tmp_path, diskfmt, diskfmt2):
     assert_equal(adata2, adata1)
 
 
-@pytest.mark.parametrize("typ", [np.array, csr_matrix])
-def test_readwrite_h5ad(typ, dataset_kwargs, backing_h5ad):
-    tmpdir = tempfile.TemporaryDirectory()
-    tmpdirpth = Path(tmpdir.name)
-    mid_pth = tmpdirpth / "mid.h5ad"
+needs_zarr = pytest.mark.skipif(not find_spec("zarr"), reason="Zarr is not installed")
 
+
+@pytest.mark.parametrize("storage", ["h5ad", pytest.param("zarr", marks=[needs_zarr])])
+@pytest.mark.parametrize("typ", [np.array, csr_matrix, as_dense_dask_array])
+def test_readwrite_kitchensink(tmp_path, storage, typ, backing_h5ad, dataset_kwargs):
     X = typ(X_list)
     adata_src = ad.AnnData(X, obs=obs_dict, var=var_dict, uns=uns_dict)
-    assert not is_categorical_dtype(adata_src.obs["oanno1"])
+    assert not isinstance(adata_src.obs["oanno1"].dtype, pd.CategoricalDtype)
     adata_src.raw = adata_src
-    adata_src.write(backing_h5ad, **dataset_kwargs)
 
-    adata_mid = ad.read(backing_h5ad)
-    adata_mid.write(mid_pth, **dataset_kwargs)
-
-    adata = ad.read_h5ad(mid_pth)
-    assert is_categorical_dtype(adata.obs["oanno1"])
-    assert not is_categorical_dtype(adata.obs["oanno2"])
+    if storage == "h5ad":
+        adata_src.write(backing_h5ad, **dataset_kwargs)
+        adata_mid = ad.read_h5ad(backing_h5ad)
+        adata_mid.write(tmp_path / "mid.h5ad", **dataset_kwargs)
+        adata = ad.read_h5ad(tmp_path / "mid.h5ad")
+    else:
+        adata_src.write_zarr(tmp_path / "test_zarr_dir", chunks=True)
+        adata = ad.read_zarr(tmp_path / "test_zarr_dir")
+    assert isinstance(adata.obs["oanno1"].dtype, pd.CategoricalDtype)
+    assert not isinstance(adata.obs["oanno2"].dtype, pd.CategoricalDtype)
     assert adata.obs.index.tolist() == ["name1", "name2", "name3"]
     assert adata.obs["oanno1"].cat.categories.tolist() == ["cat1", "cat2"]
     assert adata.obs["oanno1c"].cat.categories.tolist() == ["cat1"]
-    assert is_categorical_dtype(adata.raw.var["vanno2"])
+    assert isinstance(adata.raw.var["vanno2"].dtype, pd.CategoricalDtype)
     pd.testing.assert_frame_equal(adata.obs, adata_src.obs)
     pd.testing.assert_frame_equal(adata.var, adata_src.var)
-    assert np.all(adata.var.index == adata_src.var.index)
+    assert_equal(adata.var.index, adata_src.var.index)
     assert adata.var.index.dtype == adata_src.var.index.dtype
-    assert type(adata.raw.X) is type(adata_src.raw.X)
-    assert type(adata.raw.varm) is type(adata_src.raw.varm)
-    assert np.allclose(asarray(adata.raw.X), asarray(adata_src.raw.X))
+
+    # Dev. Note:
+    # either load as same type or load the convert DaskArray to array
+    # since we tested if assigned types and loaded types are DaskArray
+    # this would also work if they work
+    assert isinstance(adata_src.raw.X, (type(adata.raw.X), DaskArray))
+    assert isinstance(
+        adata_src.uns["uns4"]["c"], (type(adata.uns["uns4"]["c"]), DaskArray)
+    )
+    assert isinstance(adata_src.varm, (type(adata.varm), DaskArray))
+
+    assert_equal(adata.raw.X, adata_src.raw.X)
     pd.testing.assert_frame_equal(adata.raw.var, adata_src.raw.var)
     assert isinstance(adata.uns["uns4"]["a"], (int, np.integer))
     assert isinstance(adata_src.uns["uns4"]["a"], (int, np.integer))
-    assert type(adata.uns["uns4"]["c"]) is type(adata_src.uns["uns4"]["c"])
     assert_equal(adata, adata_src)
 
 
-@pytest.mark.skipif(not find_spec("zarr"), reason="Zarr is not installed")
-@pytest.mark.parametrize("typ", [np.array, csr_matrix])
-def test_readwrite_zarr(typ, tmp_path):
-    X = typ(X_list)
-    adata_src = ad.AnnData(X, obs=obs_dict, var=var_dict, uns=uns_dict)
-    adata_src.raw = adata_src
-    assert not is_categorical_dtype(adata_src.obs["oanno1"])
-    adata_src.write_zarr(tmp_path / "test_zarr_dir", chunks=True)
-
-    adata = ad.read_zarr(tmp_path / "test_zarr_dir")
-    assert is_categorical_dtype(adata.obs["oanno1"])
-    assert not is_categorical_dtype(adata.obs["oanno2"])
-    assert adata.obs.index.tolist() == ["name1", "name2", "name3"]
-    assert adata.obs["oanno1"].cat.categories.tolist() == ["cat1", "cat2"]
-    assert adata.obs["oanno1c"].cat.categories.tolist() == ["cat1"]
-    assert is_categorical_dtype(adata.raw.var["vanno2"])
-    pd.testing.assert_frame_equal(adata.obs, adata_src.obs)
-    pd.testing.assert_frame_equal(adata.var, adata_src.var)
-    assert np.all(adata.var.index == adata_src.var.index)
-    assert adata.var.index.dtype == adata_src.var.index.dtype
-    assert type(adata.raw.X) is type(adata_src.raw.X)
-    assert np.allclose(asarray(adata.raw.X), asarray(adata_src.raw.X))
-    assert np.all(adata.raw.var == adata_src.raw.var)
-    assert isinstance(adata.uns["uns4"]["a"], (int, np.integer))
-    assert isinstance(adata_src.uns["uns4"]["a"], (int, np.integer))
-    assert type(adata.uns["uns4"]["c"]) is type(adata_src.uns["uns4"]["c"])
-    assert_equal(adata, adata_src)
-
-
-@pytest.mark.parametrize("typ", [np.array, csr_matrix])
+@pytest.mark.parametrize("typ", [np.array, csr_matrix, as_dense_dask_array])
 def test_readwrite_maintain_X_dtype(typ, backing_h5ad):
-    X = typ(X_list)
-    adata_src = ad.AnnData(X, dtype="int8")
+    X = typ(X_list).astype("int8")
+    adata_src = ad.AnnData(X)
     adata_src.write(backing_h5ad)
 
-    adata = ad.read(backing_h5ad)
+    adata = ad.read_h5ad(backing_h5ad)
     assert adata.X.dtype == adata_src.X.dtype
 
 
@@ -215,38 +207,36 @@ def test_maintain_layers(rw):
     assert not np.any((orig.layers["sparse"] != curr.layers["sparse"]).toarray())
 
 
-@pytest.mark.parametrize("typ", [np.array, csr_matrix])
+@pytest.mark.parametrize("typ", [np.array, csr_matrix, as_dense_dask_array])
 def test_readwrite_h5ad_one_dimension(typ, backing_h5ad):
     X = typ(X_list)
     adata_src = ad.AnnData(X, obs=obs_dict, var=var_dict, uns=uns_dict)
     adata_one = adata_src[:, 0].copy()
     adata_one.write(backing_h5ad)
-    adata = ad.read(backing_h5ad)
+    adata = ad.read_h5ad(backing_h5ad)
     assert adata.shape == (3, 1)
     assert_equal(adata, adata_one)
 
 
-@pytest.mark.parametrize("typ", [np.array, csr_matrix])
+@pytest.mark.parametrize("typ", [np.array, csr_matrix, as_dense_dask_array])
 def test_readwrite_backed(typ, backing_h5ad):
     X = typ(X_list)
     adata_src = ad.AnnData(X, obs=obs_dict, var=var_dict, uns=uns_dict)
     adata_src.filename = backing_h5ad  # change to backed mode
     adata_src.write()
 
-    adata = ad.read(backing_h5ad)
-    assert is_categorical_dtype(adata.obs["oanno1"])
-    assert not is_categorical_dtype(adata.obs["oanno2"])
+    adata = ad.read_h5ad(backing_h5ad)
+    assert isinstance(adata.obs["oanno1"].dtype, pd.CategoricalDtype)
+    assert not isinstance(adata.obs["oanno2"].dtype, pd.CategoricalDtype)
     assert adata.obs.index.tolist() == ["name1", "name2", "name3"]
     assert adata.obs["oanno1"].cat.categories.tolist() == ["cat1", "cat2"]
     assert_equal(adata, adata_src)
 
 
 @pytest.mark.parametrize("typ", [np.array, csr_matrix, csc_matrix])
-def test_readwrite_equivalent_h5ad_zarr(typ):
-    tmpdir = tempfile.TemporaryDirectory()
-    tmpdirpth = Path(tmpdir.name)
-    h5ad_pth = tmpdirpth / "adata.h5ad"
-    zarr_pth = tmpdirpth / "adata.zarr"
+def test_readwrite_equivalent_h5ad_zarr(tmp_path, typ):
+    h5ad_pth = tmp_path / "adata.h5ad"
+    zarr_pth = tmp_path / "adata.zarr"
 
     M, N = 100, 101
     adata = gen_adata((M, N), X_type=typ)
@@ -260,6 +250,42 @@ def test_readwrite_equivalent_h5ad_zarr(typ):
     assert_equal(from_h5ad, from_zarr, exact=True)
 
 
+@contextmanager
+def store_context(path: Path):
+    if path.suffix == ".zarr":
+        store = zarr.open(path, "r+")
+    else:
+        file = h5py.File(path, "r+")
+        store = file["/"]
+    yield store
+    if "file" in locals():
+        file.close()
+
+
+@pytest.mark.parametrize(
+    ["name", "read", "write"],
+    [
+        ("adata.h5ad", ad.read_h5ad, ad.AnnData.write_h5ad),
+        ("adata.zarr", ad.read_zarr, ad.AnnData.write_zarr),
+    ],
+)
+def test_read_full_io_error(tmp_path, name, read, write):
+    adata = gen_adata((4, 3))
+    path = tmp_path / name
+    write(adata, path)
+    with store_context(path) as store:
+        store["obs"].attrs["encoding-type"] = "invalid"
+    with pytest_8_raises(
+        IORegistryError,
+        match=r"raised while reading key 'obs'.*from /$",
+    ) as exc_info:
+        read(path)
+    assert re.search(
+        r"No read method registered for IOSpec\(encoding_type='invalid', encoding_version='0.2.0'\)",
+        str(exc_info.value),
+    )
+
+
 @pytest.mark.parametrize(
     "compression,compression_opts",
     [
@@ -270,7 +296,7 @@ def test_readwrite_equivalent_h5ad_zarr(typ):
     ],
 )
 def test_hdf5_compression_opts(tmp_path, compression, compression_opts):
-    # https://github.com/theislab/anndata/issues/497
+    # https://github.com/scverse/anndata/issues/497
     pth = Path(tmp_path) / "adata.h5ad"
     adata = gen_adata((10, 8))
     kwargs = {}
@@ -299,7 +325,8 @@ def test_hdf5_compression_opts(tmp_path, compression, compression_opts):
         msg = "\n\t".join(not_compressed)
         raise AssertionError(f"These elements were not compressed correctly:\n\t{msg}")
 
-    assert_equal(adata, ad.read_h5ad(pth))
+    expected = ad.read_h5ad(pth)
+    assert_equal(adata, expected)
 
 
 def test_zarr_compression(tmp_path):
@@ -324,7 +351,8 @@ def test_zarr_compression(tmp_path):
         msg = "\n\t".join(not_compressed)
         raise AssertionError(f"These elements were not compressed correctly:\n\t{msg}")
 
-    assert_equal(adata, ad.read_zarr(pth))
+    expected = ad.read_zarr(pth)
+    assert_equal(adata, expected)
 
 
 def test_changed_obs_var_names(tmp_path, diskfmt):
@@ -363,7 +391,14 @@ def test_readwrite_loom(typ, obsm_mapping, varm_mapping, tmp_path):
     adata_src.obsm["X_a"] = np.zeros((adata_src.n_obs, 2))
     adata_src.varm["X_b"] = np.zeros((adata_src.n_vars, 3))
 
-    adata_src.write_loom(tmp_path / "test.loom", write_obsm_varm=True)
+    with warnings.catch_warnings():
+        warnings.filterwarnings("ignore", category=NumbaDeprecationWarning)
+        warnings.filterwarnings(
+            "ignore",
+            message=r"datetime.datetime.utcnow\(\) is deprecated",
+            category=DeprecationWarning,
+        )
+        adata_src.write_loom(tmp_path / "test.loom", write_obsm_varm=True)
 
     adata = ad.read_loom(
         tmp_path / "test.loom",
@@ -397,7 +432,15 @@ def test_readwrite_loom(typ, obsm_mapping, varm_mapping, tmp_path):
 def test_readloom_deprecations(tmp_path):
     loom_pth = tmp_path / "test.loom"
     adata_src = gen_adata((5, 10), obsm_types=[np.ndarray], varm_types=[np.ndarray])
-    adata_src.write_loom(loom_pth, write_obsm_varm=True)
+
+    with warnings.catch_warnings():
+        warnings.filterwarnings("ignore", category=NumbaDeprecationWarning)
+        warnings.filterwarnings(
+            "ignore",
+            message=r"datetime.datetime.utcnow\(\) is deprecated",
+            category=DeprecationWarning,
+        )
+        adata_src.write_loom(loom_pth, write_obsm_varm=True)
 
     # obsm_names -> obsm_mapping
     obsm_mapping = {"df": adata_src.obs.columns}
@@ -405,7 +448,7 @@ def test_readloom_deprecations(tmp_path):
         depr_result = ad.read_loom(loom_pth, obsm_names=obsm_mapping)
     actual_result = ad.read_loom(loom_pth, obsm_mapping=obsm_mapping)
     assert_equal(actual_result, depr_result)
-    with pytest.raises(ValueError, match="ambiguous"):
+    with pytest.raises(ValueError, match="ambiguous"), pytest.warns(FutureWarning):
         ad.read_loom(loom_pth, obsm_mapping=obsm_mapping, obsm_names=obsm_mapping)
 
     # varm_names -> varm_mapping
@@ -414,7 +457,7 @@ def test_readloom_deprecations(tmp_path):
         depr_result = ad.read_loom(loom_pth, varm_names=varm_mapping)
     actual_result = ad.read_loom(loom_pth, varm_mapping=varm_mapping)
     assert_equal(actual_result, depr_result)
-    with pytest.raises(ValueError, match="ambiguous"):
+    with pytest.raises(ValueError, match="ambiguous"), pytest.warns(FutureWarning):
         ad.read_loom(loom_pth, varm_mapping=varm_mapping, varm_names=varm_mapping)
 
     # positional -> keyword
@@ -455,12 +498,12 @@ def test_write_csv(typ, tmp_path):
 
 @pytest.mark.parametrize("typ", [np.array, csr_matrix])
 def test_write_csv_view(typ, tmp_path):
-    # https://github.com/theislab/anndata/issues/401
+    # https://github.com/scverse/anndata/issues/401
     import hashlib
 
     def md5_path(pth: PathLike) -> bytes:
         checksum = hashlib.md5()
-        with open(pth, "rb") as f:
+        with pth.open("rb") as f:
             while True:
                 buf = f.read(checksum.block_size * 100)
                 if not buf:
@@ -468,7 +511,7 @@ def test_write_csv_view(typ, tmp_path):
                 checksum.update(buf)
         return checksum.digest()
 
-    def hash_dir_contents(dir: Path) -> "dict[str, bytes]":
+    def hash_dir_contents(dir: Path) -> dict[str, bytes]:
         root_pth = str(dir)
         return {
             str(k)[len(root_pth) :]: md5_path(k) for k in dir.rglob("*") if k.is_file()
@@ -488,7 +531,7 @@ def test_write_csv_view(typ, tmp_path):
 @pytest.mark.parametrize(
     ["read", "write", "name"],
     [
-        pytest.param(ad.read_h5ad, ad._io.write._write_h5ad, "test_empty.h5ad"),
+        pytest.param(ad.read_h5ad, ad._io.write_h5ad, "test_empty.h5ad"),
         pytest.param(
             ad.read_loom,
             ad._io.write_loom,
@@ -496,17 +539,9 @@ def test_write_csv_view(typ, tmp_path):
             marks=pytest.mark.xfail(reason="Loom can’t handle 0×0 matrices"),
         ),
         pytest.param(ad.read_zarr, ad._io.write_zarr, "test_empty.zarr"),
-        pytest.param(
-            ad.read_zarr,
-            ad._io.write_zarr,
-            "test_empty.zip",
-            marks=pytest.mark.xfail(reason="Zarr zip storage doesn’t seem to work…"),
-        ),
     ],
 )
-def test_readwrite_hdf5_empty(read, write, name, tmp_path):
-    if read is ad.read_zarr:
-        pytest.importorskip("zarr")
+def test_readwrite_empty(read, write, name, tmp_path):
     adata = ad.AnnData(uns=dict(empty=np.array([], dtype=float)))
     write(tmp_path / name, adata)
     ad_read = read(tmp_path / name)
@@ -514,7 +549,13 @@ def test_readwrite_hdf5_empty(read, write, name, tmp_path):
 
 
 def test_read_excel():
-    adata = ad.read_excel(HERE / "data/excel.xlsx", "Sheet1", dtype=int)
+    with warnings.catch_warnings():
+        warnings.filterwarnings(
+            "ignore",
+            message=r"datetime.datetime.utcnow\(\) is deprecated",
+            category=DeprecationWarning,
+        )
+        adata = ad.read_excel(HERE / "data/excel.xlsx", "Sheet1", dtype=int)
     assert adata.X.tolist() == X_list
 
 
@@ -546,7 +587,7 @@ def test_write_categorical(tmp_path, diskfmt):
 def test_write_categorical_index(tmp_path, diskfmt):
     adata_pth = tmp_path / f"adata.{diskfmt}"
     orig = ad.AnnData(
-        uns={"df": pd.DataFrame(index=pd.Categorical(list("aabcd")))},
+        uns={"df": pd.DataFrame({}, index=pd.Categorical(list("aabcd")))},
     )
     getattr(orig, f"write_{diskfmt}")(adata_pth)
     curr = getattr(ad, f"read_{diskfmt}")(adata_pth)
@@ -566,17 +607,17 @@ def test_dataframe_reserved_columns(tmp_path, diskfmt):
     for colname in reserved:
         to_write = orig.copy()
         to_write.obs[colname] = np.ones(5)
-        with pytest.raises(ValueError) as e:
+        with pytest.raises(ValueError) as exc_info:
             getattr(to_write, f"write_{diskfmt}")(adata_pth)
-        assert colname in str(e.value)
+        assert colname in str(exc_info.value)
     for colname in reserved:
         to_write = orig.copy()
         to_write.varm["df"] = pd.DataFrame(
             {colname: list("aabcd")}, index=to_write.var_names
         )
-        with pytest.raises(ValueError) as e:
+        with pytest.raises(ValueError) as exc_info:
             getattr(to_write, f"write_{diskfmt}")(adata_pth)
-        assert colname in str(e.value)
+        assert colname in str(exc_info.value)
 
 
 def test_write_large_categorical(tmp_path, diskfmt):
@@ -609,7 +650,7 @@ def test_write_large_categorical(tmp_path, diskfmt):
 
 
 def test_write_string_types(tmp_path, diskfmt):
-    # https://github.com/theislab/anndata/issues/456
+    # https://github.com/scverse/anndata/issues/456
     adata_pth = tmp_path / f"adata.{diskfmt}"
 
     adata = ad.AnnData(
@@ -630,9 +671,11 @@ def test_write_string_types(tmp_path, diskfmt):
 
     adata.obs[b"c"] = np.zeros(3)
     # This should error, and tell you which key is at fault
-    with pytest.raises(TypeError, match=str(b"c")):
+    with pytest_8_raises(TypeError, match=r"writing key 'obs'") as exc_info:
         write(adata_pth)
 
+    assert "b'c'" in str(exc_info.value)
+
 
 @pytest.mark.parametrize(
     "teststring",
@@ -703,10 +746,13 @@ def test_scanpy_krumsiek11(tmp_path, diskfmt):
     filepth = tmp_path / f"test.{diskfmt}"
     import scanpy as sc
 
-    orig = sc.datasets.krumsiek11()
+    # TODO: this should be fixed in scanpy instead
+    with pytest.warns(UserWarning, match=r"Observation names are not unique"):
+        orig = sc.datasets.krumsiek11()
     del orig.uns["highlights"]  # Can’t write int keys
     getattr(orig, f"write_{diskfmt}")(filepth)
-    read = getattr(ad, f"read_{diskfmt}")(filepth)
+    with pytest.warns(UserWarning, match=r"Observation names are not unique"):
+        read = getattr(ad, f"read_{diskfmt}")(filepth)
 
     assert_equal(orig, read, exact=True)
 
@@ -765,7 +811,7 @@ def test_io_dtype(tmp_path, diskfmt, dtype):
     read = lambda pth: getattr(ad, f"read_{diskfmt}")(pth)
     write = lambda adata, pth: getattr(adata, f"write_{diskfmt}")(pth)
 
-    orig = ad.AnnData(np.ones((5, 8), dtype=dtype), dtype=dtype)
+    orig = ad.AnnData(np.ones((5, 8), dtype=dtype))
     write(orig, pth)
     curr = read(pth)
 
