@@ -1,16 +1,16 @@
 """Tests for the attribute .X"""
+from __future__ import annotations
+
 import numpy as np
 import pandas as pd
+import pytest
 from scipy import sparse
 
 import anndata as ad
 from anndata import AnnData
+from anndata.tests.helpers import assert_equal, gen_adata
 from anndata.utils import asarray
 
-import pytest
-
-from anndata.tests.helpers import gen_adata, assert_equal
-
 UNLABELLED_ARRAY_TYPES = [
     pytest.param(sparse.csr_matrix, id="csr"),
     pytest.param(sparse.csc_matrix, id="csc"),
@@ -30,7 +30,7 @@ def diskfmt(request):
 @pytest.mark.parametrize("orig_array_type", UNLABELLED_ARRAY_TYPES)
 @pytest.mark.parametrize("new_array_type", UNLABELLED_ARRAY_TYPES)
 def test_setter_singular_dim(shape, orig_array_type, new_array_type):
-    # https://github.com/theislab/anndata/issues/500
+    # https://github.com/scverse/anndata/issues/500
     adata = gen_adata(shape, X_type=orig_array_type)
     adata.X = new_array_type(np.ones(shape))
     np.testing.assert_equal(asarray(adata.X), 1)
@@ -65,8 +65,26 @@ def test_del_set_equiv_X():
     assert orig.X is None
 
 
-def test_init_X_as_none():
-    # test initialiser
+@pytest.mark.parametrize(
+    ("obs", "var", "shape_expected"),
+    [
+        pytest.param(dict(obs_names=["1", "2"]), None, (2, 0), id="obs"),
+        pytest.param(None, dict(var_names=["a", "b"]), (0, 2), id="var"),
+        pytest.param(
+            dict(obs_names=["1", "2", "3"]),
+            dict(var_names=["a", "b"]),
+            (3, 2),
+            id="both",
+        ),
+    ],
+)
+def test_init_x_as_none_shape_from_obs_var(obs, var, shape_expected):
+    adata = AnnData(None, obs, var)
+    assert adata.X is None
+    assert adata.shape == shape_expected
+
+
+def test_init_x_as_none_explicit_shape():
     shape = (3, 5)
     adata = AnnData(None, uns=dict(test=np.array((3, 3))), shape=shape)
     assert adata.X is None
